"""Asset of type IP.
This module takes care of preparing an IP asset, either single address, a range with mask for both v4 and v6."""

import ipaddress
import logging
from typing import List

import click

from ostorlab.assets import ipv4
from ostorlab.assets import ipv6
from ostorlab.cli import console as cli_console
from ostorlab.cli.scan.run import run
from ostorlab import exceptions

logger = logging.getLogger(__name__)

console = cli_console.Console()


@run.run.command(name="ip")
@click.argument("ips", required=True, nargs=-1)
@click.pass_context
def ip_cli(ctx: click.core.Context, ips: List[str]) -> None:
    """Run scan for IP asset."""
    runtime = ctx.obj["runtime"]
    try:
        assets = []
        for ip in ips:
            ip_network = ipaddress.ip_network(ip, strict=False)
            if ip_network.version == 4:
                assets.append(
                    ipv4.IPv4(
                        host=ip_network.network_address.exploded,
                        mask=str(ip_network.prefixlen),
                    )
                )
            elif ip_network.version == 6:
                assets.append(
                    ipv6.IPv6(
                        host=ip_network.network_address.exploded,
                        mask=str(ip_network.prefixlen),
                    )
                )
            else:
                console.error(f"Invalid Ip address {ip}")

        logger.debug("scanning assets %s", assets)
        try:
            created_scan = runtime.scan(
                title=ctx.obj["title"],
                agent_group_definition=ctx.obj["agent_group_definition"],
                assets=assets,
            )
            if created_scan is not None:
                runtime.link_agent_group_scan(
                    created_scan, ctx.obj["agent_group_definition"]
                )
                runtime.link_assets_scan(created_scan.id, assets)

        except exceptions.OstorlabError as e:
            console.error(f"An error was encountered while running the scan: {e}")

    except ValueError as e:
        console.error(f"{e}")
        raise click.Abort()



"""Agent asset."""

import dataclasses
from typing import Optional

from ostorlab.assets import asset


@dataclasses.dataclass
@asset.selector("v3.asset.agent")
class Agent(asset.Asset):
    """Agent asset."""

    def __init__(
        self,
        key: str,
        version: Optional[str] = None,
        git_location: Optional[str] = None,
        docker_location: Optional[str] = None,
        yaml_file_location: Optional[str] = None,
    ):
        self.key = key
        self.version = version
        self.git_location = git_location
        self.docker_location = docker_location
        self.yaml_file_location = yaml_file_location

    def __str__(self) -> str:
        if self.version is not None:
            return f"Agent {self.key}:{self.version}"
        else:
            return f"Agent {self.key}"

    @property
    def proto_field(self) -> str:
        return "agent"



"""Utils to generate and manipulate strings."""

import random
import string


def random_string(length: int, alphabet: str = string.ascii_lowercase) -> str:
    """Generates a random string of a specified length.

    Args:
        length: Length of the string to generate.
        alphabet: The alphabet used to generate the random string, defaults to ascii lower case.

    Returns:
        Randomly generated string.
    """
    if length <= 0:
        raise ValueError(f"Invalid string length: {length}")

    result = "".join((random.choice(alphabet) for _ in range(length)))
    return result



"""export utils module: contains utility functions to export the scan details."""

import io
import ipaddress
import json
import pathlib
import zipfile
from typing import Optional

from ostorlab.runtimes.local.models import models
from ostorlab.serve_app import common

SCAN_JSON = "scan.json"
ASSET_JSON = "asset.json"
VULNERABILITY_JSON = "vulnerability.json"

RISK_RATINGS_ORDER = {
    common.RiskRatingEnum.CRITICAL.name: 8,
    common.RiskRatingEnum.HIGH.name: 7,
    common.RiskRatingEnum.MEDIUM.name: 6,
    common.RiskRatingEnum.LOW.name: 5,
    common.RiskRatingEnum.POTENTIALLY.name: 4,
    common.RiskRatingEnum.HARDENING.name: 3,
    common.RiskRatingEnum.SECURE.name: 2,
    common.RiskRatingEnum.IMPORTANT.name: 1,
    common.RiskRatingEnum.INFO.name: 0,
}


def export_scan(scan: models.Scan, export_ide: bool = False) -> bytes:
    """Export the scan details from the given file data.

    Args:
        scan: The scan object.
        export_ide: Whether to export the IDE or not.
    """
    fd = io.BytesIO()
    archive = zipfile.ZipFile(fd, "a", zipfile.ZIP_DEFLATED, True)

    _export_asset(scan.id, archive)
    _export_scan(scan, archive)
    _export_vulnz(scan.id, archive)

    if export_ide is True:
        pass

    archive.close()
    fd.seek(0)
    return fd.read()


def _export_asset(scan_id: int, archive: zipfile.ZipFile) -> None:
    """Export the asset details to the given archive.

    Args:
        scan_id: The scan id.
        archive: The archive object.
    """

    with models.Database() as session:
        assets = (
            session.query(models.Asset).filter(models.Asset.scan_id == scan_id).all()
        )
        if len(assets) == 0:
            return None
        assets_data = []
        for asset in assets:
            asset_type = asset.type.replace("_", " ").lower()
            asset_dict = {
                "tags": None,
                "type": asset_type,
            }
            if asset.type == "android_file" or asset.type == "ios_file":
                path = _write_mobile_app(asset, archive, asset_type)
                if path is not None:
                    asset_dict["path"] = path

            elif asset.type == "android_store":
                asset_dict["package_name"] = asset.package_name
                asset_dict["application_name"] = asset.application_name
            elif asset.type == "ios_store":
                asset_dict["package_name"] = asset.bundle_id
                asset_dict["application_name"] = asset.application_name
            elif asset.type == "network":
                networks = _get_network(asset)
                asset_dict["networks"] = networks
            elif asset.type == "urls":
                urls = _get_urls(asset)
                asset_dict["urls"] = urls
            else:
                raise NotImplementedError(f"Asset type {asset.type} not implemented.")

            assets_data.append(asset_dict)

        assets_data = "\n".join([json.dumps(asset) for asset in assets_data])
        archive.writestr(ASSET_JSON, assets_data)


def _write_mobile_app(
    asset: models.Asset, archive: zipfile.ZipFile, asset_type: str
) -> Optional[str]:
    """Write the mobile app to the given archive.

    Args:
        asset: The asset object.
        archive: The archive object.
        asset_type: The asset type.

    Returns:
        The mobile app name.
    """

    if asset.path is None:
        return None

    mobile_app = pathlib.Path(asset.path).name
    file_asset_path = pathlib.Path(asset.path)
    if file_asset_path.exists() is False:
        raise ValueError(f"{asset_type.capitalize()} File {asset.path} not found.")
    try:
        archive.writestr(mobile_app, file_asset_path.read_bytes())
        return mobile_app
    except Exception:
        pass

    return None


def _get_network(asset: models.Asset) -> list[str]:
    """Get the network details from the given asset.

    Args:
        asset: The asset object.

    Returns:
        The network details.
    """
    with models.Database() as session:
        networks = []
        ips = (
            session.query(models.IPRange)
            .filter(models.IPRange.network_asset_id == asset.id)
            .all()
        )
        for ip in ips:
            ip_network = ipaddress.ip_network(ip.host, strict=False)
            networks.append(
                f"{ip_network.network_address.exploded}/{ip_network.prefixlen}"
            )
        return networks


def _get_urls(asset: models.Asset) -> list[str]:
    """Get the urls from the given asset.

    Args:
        asset: The asset object.

    Returns:
        The urls.
    """
    with models.Database() as session:
        urls = []
        links = (
            session.query(models.Link)
            .filter(models.Link.urls_asset_id == asset.id)
            .all()
        )
        for link in links:
            urls.append(link.url)

        return urls


def _export_scan(scan: models.Scan, archive: zipfile.ZipFile) -> None:
    """Export the scan details to the given archive.

    Args:
        scan: The scan object.
        archive: The archive object.
    """

    scan_dict = {
        "title": scan.title,
        "created_time": scan.created_time.strftime("%Y-%m-%d %H:%M:%S"),
        "risk_rating": scan.risk_rating.name.lower()
        if scan.risk_rating is not None
        else None,
        "status": [],
    }

    with models.Database() as session:
        scan_statuses = (
            session.query(models.ScanStatus)
            .filter(models.ScanStatus.scan_id == scan.id)
            .all()
        )
        if len(scan_statuses) == 0:
            scan_statuses = [
                models.ScanStatus.create(
                    key="progress", value=scan.progress.name.lower(), scan_id=scan.id
                )
            ]
        for status in scan_statuses:
            scan_dict["status"].append(
                {
                    "id": status.id,
                    "key": status.key,
                    "value": status.value,
                }
            )
        archive.writestr(SCAN_JSON, json.dumps(scan_dict))


def _export_vulnz(scan_id: int, archive: zipfile.ZipFile) -> None:
    """Export the vulnerabilities details to the given archive.

    Args:
        scan_id: The scan id.
        archive: The archive object.
    """

    with models.Database() as session:
        vulns_list = []
        vulnerabilities = session.query(models.Vulnerability).filter(
            models.Vulnerability.scan_id == scan_id
        )
        for vuln in vulnerabilities:
            references = (
                session.query(models.Reference)
                .filter(models.Reference.vulnerability_id == vuln.id)
                .all()
            )
            refs_list = [{"title": ref.title, "url": ref.url} for ref in references]
            kb_dict = {
                "title": vuln.title,
                "short_description": vuln.short_description,
                "description": vuln.description,
                "recommendation": vuln.recommendation,
                "risk_rating": vuln.risk_rating.name.lower(),
                "references": refs_list,
            }

            vulns_list.append(
                {
                    "detail": kb_dict,
                    "technical_detail": vuln.technical_detail,
                    "risk_rating": vuln.risk_rating.name.lower(),
                    "cvss_v3_vector": vuln.cvss_v3_vector,
                }
            )
        archive.writestr(VULNERABILITY_JSON, json.dumps(vulns_list))



"""Add model for ip and link

Revision ID: 5629037e8282
Revises: e0ef8ab7d038
Create Date: 2024-06-12 10:06:09.222596

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "5629037e8282"
down_revision = "e0ef8ab7d038"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "ip",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("host", sa.String(length=50), nullable=True),
        sa.Column("mask", sa.String(length=50), nullable=True),
        sa.Column("network_asset_id", sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(
            ["network_asset_id"],
            ["network.id"],
            name=op.f("fk_ip_network_asset_id_network"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_ip")),
    )
    op.create_table(
        "link",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("url", sa.String(length=4096), nullable=True),
        sa.Column("method", sa.String(length=8), nullable=True),
        sa.Column("urls_asset_id", sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(
            ["urls_asset_id"], ["urls.id"], name=op.f("fk_link_urls_asset_id_urls")
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_link")),
    )

    with op.batch_alter_table("network", schema=None) as batch_op:
        batch_op.drop_column("networks")

    with op.batch_alter_table("urls", schema=None) as batch_op:
        batch_op.drop_column("links")

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("urls", schema=None) as batch_op:
        batch_op.add_column(sa.Column("links", sa.VARCHAR(length=1024), nullable=True))

    with op.batch_alter_table("network", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column("networks", sa.VARCHAR(length=1024), nullable=True)
        )

    op.drop_table("link")
    op.drop_table("ip")
    # ### end Alembic commands ###



"""This module triggers a scan using the ios_testflight asset."""

import logging

import click

from ostorlab.cli.scan.run import run
from ostorlab.cli import console as cli_console
from ostorlab.assets import ios_testflight as ios_testflight_asset
from ostorlab import exceptions

console = cli_console.Console()
logger = logging.getLogger(__name__)


@run.run.command(name="ios-testflight")
@click.option("--application-url", required=False)
@click.pass_context
def ios_testflight(ctx: click.core.Context, application_url: str) -> None:
    """Run scan using testflight url."""
    runtime = ctx.obj["runtime"]
    assets = [ios_testflight_asset.IOSTestflight(application_url=application_url)]
    logger.debug("scanning assets %s", [str(asset) for asset in assets])
    try:
        runtime.scan(
            title=ctx.obj["title"],
            agent_group_definition=ctx.obj["agent_group_definition"],
            assets=assets,
        )
    except exceptions.OstorlabError as e:
        console.error(f"An error was encountered while running the scan: {e}")



"""Add scan risk rating

Revision ID: d47f712dab53
Revises: cbb2e16e146a
Create Date: 2024-07-12 16:49:37.847290

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "d47f712dab53"
down_revision = "cbb2e16e146a"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("scan", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "risk_rating",
                sa.Enum(
                    "CRITICAL",
                    "HIGH",
                    "MEDIUM",
                    "LOW",
                    "POTENTIALLY",
                    "HARDENING",
                    "SECURE",
                    "IMPORTANT",
                    "INFO",
                    name="riskrating",
                ),
                nullable=True,
            )
        )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("scan", schema=None) as batch_op:
        batch_op.drop_column("risk_rating")
    # ### end Alembic commands ###



"""Agent Build commands."""
import io
import logging
import pathlib

import click
import docker
from docker import errors

from ostorlab.agent.schema import loader
from ostorlab.agent.schema import validator
from ostorlab.cli import console as cli_console
from ostorlab.cli import docker_requirements_checker
from ostorlab.cli.agent import agent
from ostorlab.cli.agent.build import build_progress

console = cli_console.Console()

logger = logging.getLogger(__name__)


def _image_exists(image: str) -> bool:
    """Checks if container images exists."""
    try:
        docker_sdk_client = docker.from_env()
        docker_sdk_client.images.get(image)
        return True
    except docker.errors.ImageNotFound:
        return False


def _build_image(
    agent_name: str,
    container_name: str,
    dockerfile_path: str,
    docker_build_root: str,
    agent_definition_file: io.FileIO,
    no_cache: bool,
) -> None:
    """Build agent image from agent settings."""
    console.info(
        f"Building agent [bold red]{agent_name}[/] dockerfile [bold red]{dockerfile_path}[/]"
        f" at root [bold red]{docker_build_root}[/]."
    )
    with console.status(f"Building [bold red]{container_name}[/]"):
        for log in build_progress.BuildProgress().build(
            path=docker_build_root,
            dockerfile=dockerfile_path,
            tag=container_name,
            labels={"agent_definition": agent_definition_file.read().decode("utf-8")},
            nocache=no_cache,
        ):
            if "stream" in log and log["stream"] != "\n":
                console.info(log["stream"][:-1])
            elif "error" in log:
                console.error(log["error"][:-1])
            else:
                logger.debug(log)

    console.success(
        f"Agent {agent_name} built, container [bold red]{container_name}[/] created."
    )


@agent.command()
@click.option(
    "--file",
    "-f",
    type=click.File("rb"),
    help="Path to Agent yaml definition.",
    required=True,
)
@click.option(
    "--organization", "-o", help="Organization name.", required=False, default=""
)
@click.option("--force/--no-force", default=False)
@click.option("--no-cache", is_flag=True, default=False)
def build(
    file: io.FileIO, organization: str = "", force: bool = False, no_cache: bool = False
) -> None:
    """CLI command to build the agent container from a definition.yaml file.
    Usage : Ostorlab agent build -f path/to/definition.yaml -org organization_name
    """

    if not docker_requirements_checker.is_docker_installed():
        console.error("Docker is not installed.")
        raise click.exceptions.Exit(2)
    if not docker_requirements_checker.is_user_permitted():
        console.error("User does not have permissions to run docker.")
        raise click.exceptions.Exit(2)
    if not docker_requirements_checker.is_docker_working():
        console.error("Error using docker.")
        raise click.exceptions.Exit(2)

    try:
        agent_def = loader.load_agent_yaml(file)
        file.seek(0)
        dockerfile_path = agent_def["docker_file_path"]
        docker_build_root = agent_def["docker_build_root"]

        _check_build_root(docker_build_root, file)

        agent_name = agent_def["name"]
        agent_version = agent_def.get("version", "0.0.0")
        container_name = f"agent_{organization}_{agent_name}:v{agent_version}"
        client = docker.from_env()

        if _image_exists(container_name):
            if force is False:
                console.info(f"{container_name} already exist.")
                raise click.exceptions.Exit(0)
            else:
                console.warning(f"{container_name} already exist, deleting image.")
                client.images.remove(container_name, force=True)
        _build_image(
            agent_name,
            container_name,
            dockerfile_path,
            docker_build_root,
            file,
            no_cache,
        )

    except errors.BuildError as e:
        console.error("Error building agent.")
        raise click.exceptions.Exit(2) from e
    except validator.SchemaError as e:
        console.error(
            "Schema is invalid, this should not happen, please report an issue at "
            "https://github.com/Ostorlab/ostorlab/issues."
        )
        raise click.exceptions.Exit(2) from e
    except validator.ValidationError as e:
        console.error(f"Definition file does not conform to the provided specification: {e}")
        raise click.exceptions.Exit(2) from e


def _check_build_root(docker_build_root, file):
    """Check whether build root is valid.

    Checks if docker build root is a subfolder of yaml file path.
    """
    yaml_file_dir = pathlib.Path(file.name).parent.resolve()
    docker_build_root_path = (yaml_file_dir / docker_build_root).resolve()
    if str(yaml_file_dir) not in str(docker_build_root_path):
        console.error(f"Invalid docker build path {docker_build_root}.")
        raise click.exceptions.Exit(3)



"""Module responsible for installing an agent : Pulling the image from the ostorlab store."""

import logging
from typing import Dict, Optional, Generator

import click
import docker
import docker.errors

from ostorlab.cli import console as cli_console
from ostorlab.cli.agent.install import install_progress
from ostorlab.cli import agent_fetcher

logger = logging.getLogger(__name__)

console = cli_console.Console()


def _image_name_from_key(agent_key: str) -> str:
    """Generate docker image name from an agent key.

    Args:
        agent_key: agent key in the form agent/organization/name.

    Returns:
        image_name: image name in the form : agent_organization_name.
    """
    return agent_key.replace("/", "_").lower()


def _parse_repository_tag(repo_name: str, tag: str = None) -> tuple:
    """Get repo name and tag"""
    parts = repo_name.rsplit("@", 1)
    if len(parts) == 2:
        return tuple(parts)
    parts = repo_name.rsplit(":", 1)
    if len(parts) == 2 and "/" not in parts[1]:
        return tuple(parts)
    return repo_name, tag


def _pull_logs(
    docker_client: docker.DockerClient, repository: str, tag: Optional[str] = None
) -> Generator[Dict, None, None]:
    """Generate logs of the docker pull method."""
    repository, tag = _parse_repository_tag(repository, tag)
    pull_log = docker_client.api.pull(repository, tag=tag, stream=True, decode=True)
    for log in pull_log:
        yield log


def _get_image(
    docker_client: docker.DockerClient, repository: str, tag: str = None
) -> docker.models.images.Image:
    """Gets a docker image."""
    repository, tag = _parse_repository_tag(repository, tag)
    name = f"{repository}:{tag}"

    return docker_client.images.get(name)


def _is_image_present(docker_client: docker.DockerClient, image_name: str) -> bool:
    """Check if a docker image exists."""
    try:
        docker_client.images.get(image_name)
        return True
    except docker.errors.ImageNotFound:
        return False


def install(
    agent_key: str,
    version: str = "",
    docker_client: Optional[docker.DockerClient] = None,
) -> None:
    """Install an agent : Fetch the docker file location of the agent corresponding to the agent_key,
    and pull the image from the registry.

    Args:
        agent_key: key of the agent in agent/org/agentName format.
        version: version of the docker image.
        docker_client: optional instance of the docker client to use to install the agent.

    Returns:
        None

    Raises:
        click Exit exception with status code 2 when the docker image does not exist.
    """

    agent_details = agent_fetcher.get_details(agent_key)
    agent_docker_location = agent_details["dockerLocation"]
    if agent_docker_location is None or not agent_details.get("versions", {}).get(
        "versions", []
    ):
        console.error(f"Agent: {agent_key} image location is not yet available")
        raise click.exceptions.Exit(2)

    image_name = _image_name_from_key(agent_details["key"])
    expected_version = version or agent_details["versions"]["versions"][0]["version"]
    logger.debug("searching for image name %s", image_name)

    try:
        docker_client = docker_client or docker.from_env()

        if _is_image_present(docker_client, f"{image_name}:v{expected_version}"):
            console.info(f"{agent_key} already exist.")
        else:
            console.info(
                f"Pulling the image {agent_docker_location} from the ostorlab store."
            )
            pull_logs_generator = _pull_logs(
                docker_client, agent_docker_location, f"v{expected_version}"
            )

            agent_install_progress = install_progress.AgentInstallProgress()
            agent_install_progress.display(pull_logs_generator)

            agent_image = _get_image(
                docker_client=docker_client,
                repository=agent_docker_location,
                tag=f"v{expected_version}",
            )
            agent_image.tag(repository=image_name, tag=f"v{expected_version}")

    except docker.errors.ImageNotFound as e:
        error_message = f"Image of the provided agent : {agent_key} was not found."
        console.error(error_message)
        raise click.exceptions.Exit(2) from e



"""Add reference table

Revision ID: 23afcfd1d789
Revises: cff0aef2c0f7
Create Date: 2024-05-30 18:29:17.581830

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "23afcfd1d789"
down_revision = "cff0aef2c0f7"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "reference",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("title", sa.String(length=255), nullable=True),
        sa.Column("url", sa.String(length=4096), nullable=True),
        sa.Column("vulnerability_id", sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(
            ["vulnerability_id"],
            ["vulnerability.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    with op.batch_alter_table("agent_argument", schema=None) as batch_op:
        batch_op.alter_column(
            "value", existing_type=sa.BLOB(), type_=sa.Text(), existing_nullable=True
        )

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("agent_argument", schema=None) as batch_op:
        batch_op.alter_column(
            "value", existing_type=sa.Text(), type_=sa.BLOB(), existing_nullable=True
        )

    op.drop_table("reference")
    # ### end Alembic commands ###



"""Module contains all the supported assets that we can pass to the scan command."""

from ostorlab.cli.scan.run.assets import agent
from ostorlab.cli.scan.run.assets import android_aab
from ostorlab.cli.scan.run.assets import android_apk
from ostorlab.cli.scan.run.assets import domain_name
from ostorlab.cli.scan.run.assets import file
from ostorlab.cli.scan.run.assets import ios_ipa
from ostorlab.cli.scan.run.assets import ip
from ostorlab.cli.scan.run.assets import link
from ostorlab.cli.scan.run.assets import android_store
from ostorlab.cli.scan.run.assets import ios_store
from ostorlab.cli.scan.run.assets import ios_testflight
from ostorlab.cli.scan.run.assets import api_schema

__all__ = (
    "agent",
    "android_aab",
    "android_apk",
    "domain_name",
    "file",
    "ios_ipa",
    "ip",
    "link",
    "android_store",
    "ios_store",
    "ios_testflight",
    "api_schema",
)



"""This module triggers a scan using the ApiSchema asset."""

import io
import logging
from typing import Optional

import click

from ostorlab.assets import api_schema as api_schema_asset
from ostorlab.cli import console as cli_console
from ostorlab.cli.scan.run import run
from ostorlab import exceptions

console = cli_console.Console()
logger = logging.getLogger(__name__)


@run.run.command(name="api-schema")
@click.option("--url", help="The URL to scan.", multiple=False, required=True)
@click.option(
    "--schema-file",
    type=click.File(mode="rb"),
    help="The file path of the schema.",
    multiple=False,
    required=False,
)
@click.option(
    "--schema-url",
    help="The URL from which to download the schema.",
    multiple=False,
    required=False,
)
@click.option(
    "--schema-type",
    help="The schema type (graphql, wsdl, or openapi).",
    multiple=False,
    required=False,
)
@click.pass_context
def api_schema(
    ctx: click.core.Context,
    url: str,
    schema_file: Optional[io.FileIO] = None,
    schema_url: Optional[str] = None,
    schema_type: Optional[str] = None,
) -> None:
    """Run scan for an API Schema asset.

    Args:
        ctx: The Click context.
        url: The URL to scan.
        schema_file: The file path of the schema.
        schema_url: The URL from which to download the schema.
        schema_type: Optional type of schema (graphql, wsdl, openapi).
    """
    if schema_file is None and schema_url is None:
        console.error("You must provide either --schema-file or --schema-url.")
        raise click.exceptions.Exit(2)
    runtime = ctx.obj["runtime"]
    assets = [
        api_schema_asset.ApiSchema(
            content=schema_file.read() if schema_file is not None else None,
            content_url=schema_url,
            endpoint_url=url,
            schema_type=schema_type,
        )
    ]

    logger.debug("Scanning assets %s", [str(asset) for asset in assets])
    try:
        runtime.scan(
            title=ctx.obj["title"],
            agent_group_definition=ctx.obj["agent_group_definition"],
            assets=assets,
        )
    except exceptions.OstorlabError as e:
        console.error(f"An error was encountered while running the scan: {e}")



"""Create mobile scan API."""

import enum
import io
import json
from typing import Dict, Optional, BinaryIO, List

from . import request

SCAN_PROFILES = {
    "fast_scan": "Fast Scan",
    "full_scan": "Full Scan",
    # aliases
    "fast": "Fast Scan",
    "full": "Full Scan",
}


class MobileAssetType(enum.Enum):
    ANDROID = enum.auto()
    IOS = enum.auto()


class CreateMobileScanAPIRequest(request.APIRequest):
    """Create mobile scan API from a file."""

    def __init__(
        self,
        title: str,
        asset_type: MobileAssetType,
        scan_profile: str,
        application: BinaryIO,
        test_credential_ids: Optional[List[int]] = None,
        sboms: list[io.FileIO] = None,
    ):
        self._title = title
        self._asset_type = asset_type
        self._scan_profile = scan_profile
        self._application = application
        self._test_credential_ids = test_credential_ids
        self._sboms = sboms

    @property
    def query(self) -> Optional[str]:
        """Defines the query to create a mobile scan.

        Returns:
            The query to create a mobile scan
        """

        return """
mutation MobileScan($title: String!, $assetType: String!, $application: Upload!, $sboms: [Upload!], $scanProfile: String!, $credentialIds: [Int]) {
  createMobileScan(title: $title, assetType: $assetType, application: $application, sboms: $sboms, scanProfile: $scanProfile, credentialIds: $credentialIds) {
    scan {
      id
    }
  }
}
        """

    @property
    def data(self) -> Optional[Dict]:
        """Sets the query and variables to create the scan.

        Returns:
            The query and variables to create a scan.
        """

        var_map = {"0": ["variables.application"]}
        for idx, _ in enumerate(self._sboms):
            var_map[str(idx + 1)] = [f"variables.sboms.{idx}"]

        data = {
            "operations": json.dumps(
                {
                    "query": self.query,
                    "variables": {
                        "title": self._title,
                        "assetType": self._asset_type.name.lower(),
                        "application": None,
                        "scanProfile": self._scan_profile,
                        "credentialIds": self._test_credential_ids,
                        "sboms": [None for _ in self._sboms],
                    },
                }
            ),
            "map": json.dumps(var_map),
        }
        return data

    @property
    def files(self) -> Optional[Dict]:
        """Sets the file for multipart upload to create the mobile scan.

        Returns:
            The file mapping to create a scan.
        """
        files = {"0": self._application}
        for idx, sbom_file in enumerate(self._sboms):
            files[str(idx + 1)] = sbom_file
        return files



"""Module for the ostorlab serve command."""

import sys

import click

from ostorlab.cli import console as cli_console
from ostorlab.cli.rootcli import rootcli
from ostorlab.runtimes.local.models import models

console = cli_console.Console()
DEFAULT_SCANNER_NAME = "Local Scanner"


@rootcli.command()
@click.option("--host", default="0.0.0.0", help="The host to run the Flask app on.")
@click.option("--port", default=3420, help="The port to run the Flask app on.")
@click.option("--refresh-api-key", is_flag=True, help="Generate a new API key.")
@click.pass_context
def serve(ctx: click.core.Context, host: str, port: int, refresh_api_key: bool) -> None:
    """Run the Flask serve with the specified host and port."""
    try:
        from ostorlab.serve_app import app
        import werkzeug

        cli = sys.modules["flask.cli"]
        cli.show_server_banner = lambda *x: None
        # Disable werkzeug log_startup, this can't be done by overriding BaseWSGIServer class because it's used by Flask.
        werkzeug.serving.BaseWSGIServer.log_startup = lambda *x: None
    except ImportError as e:
        missing_dependency = e.name
        raise click.ClickException(
            f"The '{missing_dependency}' package is required for the 'serve' command. "
            f"Please install it using 'pip install ostorlab[serve]'."
        )
    if refresh_api_key is True:
        api_key = models.APIKey.refresh()
        console.info("API key refreshed")
    else:
        api_key = models.APIKey.get_or_create()
    flask_app = app.create_app(graphiql=True)
    graphql_endpoint = f"http://{host}:{port}/graphql"
    ui_url = f"http://{host}:{port}/"
    scanners_url = f"{ui_url}scanners/new?endpoint={graphql_endpoint}&key={api_key.key}&name={DEFAULT_SCANNER_NAME}"
    console.success(f"Quick start: {scanners_url}")
    console.success(f"To authenticate, please use the following API key: {api_key.key}")
    console.success(f"Serving API on : {graphql_endpoint}")
    console.success(f"Serving UI on : {ui_url}")
    flask_app.run(host=host, port=port)



"""Ostorlab cli package."""

from ostorlab.cli import agent
from ostorlab.cli import agentgroup
from ostorlab.cli import auth
from ostorlab.cli import ci_scan
from ostorlab.cli import scan
from ostorlab.cli import scanner
from ostorlab.cli import serve
from ostorlab.cli import vulnz

__all__ = [
    "scan",
    "auth",
    "agent",
    "vulnz",
    "agentgroup",
    "ci_scan",
    "scanner",
    "serve",
]



"""Module for the command run inside the group ci_scan.
Example of usage:
- ostorlab --api-key='myKey' ci-scan run --scan-profile=full_scan \
           --break-on-risk-rating=medium --title=test_scan [asset] [options].
"""

import io
import multiprocessing
import click
import time
from typing import List

from ostorlab.cli.ci_scan.ci_scan import ci_scan
from ostorlab.apis import scan_create as scan_create_api
from ostorlab.apis import scan_info as scan_info_api
from ostorlab.apis.runners import authenticated_runner
from ostorlab.cli.ci_scan.run.ci_logger import (
    console_logger,
    github_logger,
    circleci_logger,
    logger,
)
from ostorlab.utils import risk_rating

MINUTE = 60
WAIT_MINUTES = 30
SLEEP_CHECKS = 10  # seconds
SCAN_PROGRESS_NOT_STARTED = "not_started"
SCAN_PROGRESS_DONE = "done"

CI_LOGGER = {
    "console": console_logger.Logger,
    "github": github_logger.Logger,
    "circleci": circleci_logger.Logger,
}


@ci_scan.group()
@click.option("--scan-profile", help="Scan profile to execute.", required=True)
@click.option("--title", help="Scan title.")
@click.option(
    "--break-on-risk-rating",
    help="Fail if the scan risk rating is higher than the defined value.",
)
@click.option(
    "--max-wait-minutes",
    help="Time to wait for the scan results.",
    required=False,
    default=WAIT_MINUTES,
)
@click.option(
    "--log-flavor",
    help="Type of expected output based on the CI.",
    required=False,
    default="console",
)
@click.option(
    "--test-credentials-login",
    help="Test credentials login, composed of login, password, url (optional) and role (optional).",
    required=False,
    multiple=True,
)
@click.option(
    "--test-credentials-password",
    help="Test credentials login, composed of login, password, url (optional) and role (optional).",
    required=False,
    multiple=True,
)
@click.option(
    "--test-credentials-url",
    help="Test credentials login, composed of login, password, url (optional) and role (optional).",
    required=False,
    multiple=True,
)
@click.option(
    "--test-credentials-role",
    help="Test credentials login, composed of login, password, url (optional) and role (optional).",
    required=False,
    multiple=True,
)
@click.option(
    "--test-credentials-name",
    help="Test custom credentials, composed of name and value.",
    required=False,
    multiple=True,
)
@click.option(
    "--test-credentials-value",
    help="Test custom credentials, composed of name and value.",
    required=False,
    multiple=True,
)
@click.option(
    "--sbom",
    "sboms",
    help="Path to sbom file.",
    type=click.File(mode="rb"),
    required=False,
    multiple=True,
    default=[],
)
@click.pass_context
def run(
    ctx: click.core.Context,
    scan_profile: str,
    title: str,
    break_on_risk_rating: str,
    max_wait_minutes: int,
    log_flavor: str,
    test_credentials_login: List[str],
    test_credentials_password: List[str],
    test_credentials_url: List[str],
    test_credentials_role: List[str],
    test_credentials_name: List[str],
    test_credentials_value: List[str],
    sboms: List[io.FileIO],
) -> None:
    """Start a scan based on a scan profile in the CI.\n"""

    if log_flavor not in CI_LOGGER:
        CI_LOGGER["console"]().error(
            f"log_flavor value {log_flavor} not supported."
            f" Possible options: {CI_LOGGER.keys()}"
        )
        raise click.exceptions.Exit(2)

    ci_logger = CI_LOGGER.get(log_flavor)()

    if len(test_credentials_login) != len(test_credentials_password):
        ci_logger.error("Loging and password credentials are not matching count.")
        raise click.exceptions.Exit(2)

    if len(test_credentials_name) != len(test_credentials_value):
        ci_logger.error("Name and value credentials are not matching count.")
        raise click.exceptions.Exit(2)

    if not ctx.obj.get("api_key"):
        ci_logger.error("API key not not provided.")
        raise click.exceptions.Exit(2)

    if scan_profile in scan_create_api.SCAN_PROFILES:
        ctx.obj["scan_profile"] = scan_create_api.SCAN_PROFILES[scan_profile]
    else:
        ci_logger.error(
            f"Scan profile {scan_profile} not supported. Possible options: {scan_create_api.SCAN_PROFILES.keys()}"
        )
        raise click.exceptions.Exit(2)

    ctx.obj["title"] = title
    ctx.obj["break_on_risk_rating"] = break_on_risk_rating
    ctx.obj["max_wait_minutes"] = max_wait_minutes
    ctx.obj["ci_logger"] = ci_logger
    ctx.obj["test_credentials"] = {
        "test_credentials_login": test_credentials_login,
        "test_credentials_password": test_credentials_password,
        "test_credentials_url": test_credentials_url,
        "test_credentials_role": test_credentials_role,
        "test_credentials_name": test_credentials_name,
        "test_credentials_value": test_credentials_value,
    }
    ctx.obj["sboms"] = sboms


def apply_break_scan_risk_rating(
    break_on_risk_rating: str,
    scan_id: int,
    max_wait_minutes: int,
    runner: authenticated_runner.AuthenticatedAPIRunner,
    ci_logger: logger.Logger,
):
    """Wait for the scan to finish and raise an exception if its risk is higher than the defined value."""
    if risk_rating.RiskRating.has_value(break_on_risk_rating.upper()):
        # run the check on a separate process and join it with a timeout. An exception is raised after the timeout.
        check_scan_process = multiprocessing.Process(
            target=_check_scan_periodically, args=(runner, scan_id)
        )
        check_scan_process.start()
        check_scan_process.join(int(max_wait_minutes) * MINUTE)

        if check_scan_process.is_alive():
            check_scan_process.kill()
            check_scan_process.join()
            _handle_scan_timeout(runner, scan_id, break_on_risk_rating, ci_logger)

        scan_result = runner.execute(scan_info_api.ScanInfoAPIRequest(scan_id=scan_id))
        if scan_result["data"]["scan"]["progress"] == "done":
            scan_risk_rating = scan_result["data"]["scan"]["riskRating"]
            _check_scan_risk_rating(scan_risk_rating, break_on_risk_rating, ci_logger)
    else:
        ci_logger.error(
            f"Incorrect risk rating value {break_on_risk_rating}. "
            f'It should be one of {", ".join(risk_rating.RiskRating.values())}'
        )
        raise click.exceptions.Exit(2)


def _check_scan_periodically(
    runner: authenticated_runner.AuthenticatedAPIRunner, scan_id: int
) -> None:
    """Retrieve the scan progress using the API and wait if the progress is different than done"""
    scan_done = False
    while not scan_done:
        scan_result = runner.execute(scan_info_api.ScanInfoAPIRequest(scan_id=scan_id))
        scan_progress = scan_result["data"]["scan"]["progress"]
        if scan_progress == "done":
            scan_done = True
        time.sleep(SLEEP_CHECKS)


def _check_scan_risk_rating(
    scan_risk_rating: str, break_on_risk_rating, ci_logger: logger.Logger
) -> None:
    """Compare the scan risk and raise an exception if the scan risk is higher than the defined value"""
    if _is_scan_risk_rating_higher(scan_risk_rating, break_on_risk_rating):
        ci_logger.error(f"The scan risk rating is {scan_risk_rating}.")
        raise click.exceptions.Exit(2)
    else:
        ci_logger.info(f"Scan done with risk rating {scan_risk_rating}.")


def _is_scan_risk_rating_higher(
    scan_risk_rating: str, break_on_risk_rating: str
) -> bool:
    """Returns a boolean of the scan risk Comparison with the defined break value"""
    return (
        scan_risk_rating.upper() in risk_rating.RATINGS_ORDER
        and risk_rating.RATINGS_ORDER[scan_risk_rating.upper()]
        < risk_rating.RATINGS_ORDER[break_on_risk_rating.upper()]
    )


def _handle_scan_timeout(
    runner: authenticated_runner.AuthenticatedAPIRunner,
    scan_id: int,
    break_on_risk_rating: str,
    ci_logger: logger.Logger,
) -> None:
    """when the scan triggers a timeout, we check if the scan has started and if the risk rating is higher
    than the defined break value. In this case we report the risk rating, and we exit code 2
    otherwise we raise the timeout exception
    """
    scan_result = runner.execute(scan_info_api.ScanInfoAPIRequest(scan_id=scan_id))
    if scan_result["data"]["scan"]["progress"] != SCAN_PROGRESS_NOT_STARTED:
        scan_risk_rating = scan_result["data"]["scan"]["riskRating"]
        if scan_risk_rating is not None and _is_scan_risk_rating_higher(
            scan_risk_rating, break_on_risk_rating
        ):
            ci_logger.error(f"The scan risk rating is {scan_risk_rating}.")
            raise click.exceptions.Exit(2)
    raise TimeoutError()



"""Reporter logic to read the scanner state periodically and send it to the backend."""

from ostorlab.apis.runners import authenticated_runner
from ostorlab.apis import add_scanner_state
from ostorlab.utils import defintions


class ScannerStateReporter:
    """Reporter collects information about the current scanner."""

    def __init__(
        self,
        scanner_id: str,
        hostname: str,
        ip: str,
    ):
        self._scanner_id = scanner_id
        self.scan_id = None
        self._hostname = hostname
        self._ip = ip

    def _capture_state(self) -> defintions.ScannerState:
        """Capture current scanner state."""
        try:
            import psutil

            state = defintions.ScannerState(
                scanner_id=self._scanner_id,
                scan_id=self.scan_id,
                cpu_load=psutil.cpu_percent(interval=1, percpu=False),
                total_cpu=psutil.cpu_count(),
                memory_load=psutil.virtual_memory().percent,
                total_memory=psutil.virtual_memory().total >> 30,  # total memory in GB
                hostname=self._hostname,
                ip=self._ip,
            )
        except ImportError:
            state = defintions.ScannerState(
                scanner_id=self._scanner_id,
                scan_id=self.scan_id,
                cpu_load=0,
                total_cpu=0,
                memory_load=0,
                total_memory=0,  # total memory in GB
                hostname=self._hostname,
                ip=self._ip,
            )

        return state

    def _report_state(self, state: defintions.ScannerState) -> None:
        runner = authenticated_runner.AuthenticatedAPIRunner()
        _ = runner.execute(add_scanner_state.AddScannerStateAPIRequest(state=state))

    async def report(self) -> None:
        """Capture the current state of the scanner and persist it."""
        state = self._capture_state()
        self._report_state(state)



"""API Schema asset."""

import dataclasses
from typing import Optional

from ostorlab.assets import asset

PROTO_FIELD = "api_schema"
CLASS_NAME = "API Schema"


@dataclasses.dataclass
@asset.selector("v3.asset.file.api_schema")
class ApiSchema(asset.Asset):
    """API Schema target asset."""

    def __init__(
        self,
        endpoint_url: str,
        content: Optional[bytes] = None,
        content_url: Optional[str] = None,
        schema_type: Optional[str] = None,
    ):
        self.content = content
        self.endpoint_url = endpoint_url
        self.schema_type = schema_type
        self.content_url = content_url

    def __str__(self) -> str:
        str_representation = CLASS_NAME
        if self.schema_type is not None:
            str_representation = f"{str_representation} ({self.schema_type})"
        str_representation = f"{str_representation}: {self.endpoint_url}"

        return str_representation

    @property
    def proto_field(self) -> str:
        return PROTO_FIELD



"""Asset of type Link."""

import logging
from typing import List

import click

from ostorlab.assets import link as link_asset
from ostorlab.cli.scan.run import run
from ostorlab.cli import console as cli_console
from ostorlab import exceptions

logger = logging.getLogger(__name__)
console = cli_console.Console()


@run.run.command()
@click.option("--url", help="List of Urls to scan.", required=True, multiple=True)
@click.option("--method", help="List of HTTP methods.", required=True, multiple=True)
@click.pass_context
def link(ctx: click.core.Context, url: List[str], method: List[str]) -> None:
    """Run scan for links."""
    runtime = ctx.obj["runtime"]
    if len(url) != len(method):
        console.error("Make sure every URL has its corresponding method.")
        raise click.exceptions.Exit(2)
    assets = []
    for u, m in zip(url, method):
        asset = link_asset.Link(url=u, method=m)
        assets.append(asset)
    logger.debug("scanning assets %s", asset)
    try:
        created_scan = runtime.scan(
            title=ctx.obj["title"],
            agent_group_definition=ctx.obj["agent_group_definition"],
            assets=assets,
        )
        if created_scan is not None:
            runtime.link_agent_group_scan(
                created_scan, ctx.obj["agent_group_definition"]
            )
            runtime.link_assets_scan(created_scan.id, assets)

    except exceptions.OstorlabError as e:
        console.error(f"An error was encountered while running the scan: {e}")



"""Agent and Agent group definitions and settings dataclasses."""

import dataclasses
import io
import pathlib
import logging
import json
from typing import List, Dict, Optional, Any
import ipaddress


from ostorlab.agent.schema import loader
from ostorlab.cli import agent_fetcher
from ostorlab.runtimes.proto import agent_instance_settings_pb2
from ostorlab.utils import defintions
from ostorlab.assets import android_apk as android_apk_asset
from ostorlab.assets import android_aab as android_aab_asset
from ostorlab.assets import android_store as android_store_asset
from ostorlab.assets import domain_name as domain_name_asset
from ostorlab.assets import ios_ipa as ios_ipa_asset
from ostorlab.assets import ios_store as ios_store_asset
from ostorlab.assets import ipv4 as ipv4_asset
from ostorlab.assets import ipv6 as ipv6_asset
from ostorlab.assets import link as link_asset
from ostorlab.assets import asset as base_asset

MAX_AGENT_REPLICAS = 100

logger = logging.getLogger(__name__)


def _process_agent_replicas(replicas: int) -> int:
    """Add an upper & lower bounds to the number of replicas allowed for an Agent instance."""
    if replicas <= 0:
        return 1
    elif replicas > MAX_AGENT_REPLICAS:
        return MAX_AGENT_REPLICAS
    else:
        return replicas


@dataclasses.dataclass
class AgentSettings:
    """Agent instance lists the settings of running instance of an agent."""

    key: str
    version: Optional[str] = None
    bus_url: Optional[str] = ""
    bus_exchange_topic: Optional[str] = ""
    bus_management_url: Optional[str] = ""
    bus_vhost: Optional[str] = ""
    args: List[defintions.Arg] = dataclasses.field(default_factory=list)
    constraints: Optional[List[str]] = dataclasses.field(default_factory=list)
    mounts: Optional[List[str]] = dataclasses.field(default_factory=list)
    restart_policy: str = ""
    mem_limit: Optional[int] = None
    open_ports: List[defintions.PortMapping] = dataclasses.field(default_factory=list)
    replicas: int = 1
    healthcheck_host: str = "0.0.0.0"
    healthcheck_port: int = 5000
    redis_url: Optional[str] = None
    tracing_collector_url: Optional[str] = None
    caps: Optional[List[str]] = None
    cyclic_processing_limit: Optional[int] = None
    depth_processing_limit: Optional[int] = None
    accepted_agents: Optional[List[str]] = None
    in_selectors: Optional[List[str]] = dataclasses.field(default_factory=list)

    @property
    def container_image(self):
        return agent_fetcher.get_container_image(self.key, self.version)

    @classmethod
    def from_proto(cls, proto: bytes) -> "AgentSettings":
        """Constructs an agent definition from a binary proto settings.

        Args:
            proto: binary proto settings file.

        Returns:
            AgentInstanceSettings object.
        """
        instance = agent_instance_settings_pb2.AgentInstanceSettings()
        instance.ParseFromString(proto)
        return cls(
            key=instance.key,
            bus_url=instance.bus_url,
            bus_exchange_topic=instance.bus_exchange_topic,
            bus_management_url=instance.bus_management_url,
            bus_vhost=instance.bus_vhost,
            args=[
                defintions.Arg(name=a.name, type=a.type, value=a.value)
                for a in instance.args
            ],
            constraints=instance.constraints,
            mounts=instance.mounts,
            restart_policy=instance.restart_policy,
            mem_limit=instance.mem_limit,
            open_ports=[
                defintions.PortMapping(
                    source_port=p.source_port, destination_port=p.destination_port
                )
                for p in instance.open_ports
            ],
            replicas=instance.replicas,
            healthcheck_host=instance.healthcheck_host,
            healthcheck_port=instance.healthcheck_port,
            redis_url=instance.redis_url,
            tracing_collector_url=instance.tracing_collector_url,
            caps=instance.caps,
            cyclic_processing_limit=instance.cyclic_processing_limit,
            depth_processing_limit=instance.depth_processing_limit,
            accepted_agents=instance.accepted_agents,
            in_selectors=instance.in_selectors,
        )

    def to_raw_proto(self) -> bytes:
        """Transforms agent instance settings into a raw proto bytes.

        Returns:
            Bytes as a serialized proto.
        """
        instance = agent_instance_settings_pb2.AgentInstanceSettings()
        instance.key = self.key
        instance.bus_url = self.bus_url
        instance.bus_exchange_topic = self.bus_exchange_topic
        instance.bus_management_url = self.bus_management_url
        instance.bus_vhost = self.bus_vhost

        for arg in self.args:
            arg_instance = instance.args.add()
            arg_instance.name = arg.name
            arg_instance.type = arg.type
            if isinstance(arg.value, bytes) and arg_instance.type != "binary":
                raise ValueError(
                    f"type {arg_instance.type} for not match value of type binary"
                )

            if isinstance(arg.value, bytes) and arg_instance.type == "binary":
                arg_instance.value = arg.value
            else:
                try:
                    arg_instance.value = json.dumps(arg.value).encode()
                except TypeError as e:
                    raise ValueError(
                        f"type {arg_instance.value} is not JSON serializable"
                    ) from e

        instance.constraints.extend(self.constraints)
        instance.mounts.extend(self.mounts)
        instance.restart_policy = self.restart_policy
        instance.in_selectors.extend(self.in_selectors)
        if self.mem_limit is not None:
            instance.mem_limit = self.mem_limit

        for open_port in self.open_ports:
            open_port_instance = instance.open_ports.add()
            open_port_instance.source_port = open_port.source_port
            open_port_instance.destination_port = open_port.destination_port

        instance.replicas = self.replicas
        instance.healthcheck_host = self.healthcheck_host
        instance.healthcheck_port = self.healthcheck_port
        instance.caps.extend(self.caps)

        if self.cyclic_processing_limit is not None:
            instance.cyclic_processing_limit = self.cyclic_processing_limit

        if self.depth_processing_limit is not None:
            instance.depth_processing_limit = self.depth_processing_limit

        if self.accepted_agents is not None:
            instance.accepted_agents.extend(self.accepted_agents)

        if self.redis_url is not None:
            instance.redis_url = self.redis_url

        if self.tracing_collector_url is not None:
            instance.tracing_collector_url = self.tracing_collector_url

        return instance.SerializeToString()


@dataclasses.dataclass
class AgentGroupDefinition:
    """Data class holding the attributes of an agent."""

    agents: List[AgentSettings]
    name: Optional[str] = None
    description: Optional[str] = None

    @classmethod
    def from_yaml(cls, group: io.FileIO):
        """Construct AgentGroupDefinition from yaml file.

        Args:
            group : agent group .yaml file.
        """
        agent_group_def = loader.load_agent_group_yaml(group)
        agent_settings = []
        agents_names = []
        for agent in agent_group_def["agents"]:
            agents_names.append(agent.get("key"))
            agent_def = AgentSettings(
                key=agent.get("key"),
                version=agent.get("version"),
                args=[
                    defintions.Arg(
                        name=a.get("name"),
                        description=a.get("description"),
                        type=a.get("type"),
                        value=a.get("value"),
                    )
                    for a in agent.get("args", [])
                ],
                constraints=agent.get("constraints", []),
                mounts=agent.get("mounts", []),
                restart_policy=agent.get("restart_policy", ""),
                mem_limit=agent.get("mem_limit"),
                open_ports=[
                    defintions.PortMapping(
                        source_port=p.get("src_port"),
                        destination_port=p.get("dest_port"),
                    )
                    for p in agent.get("open_ports", [])
                ],
                replicas=agent.get("replicas", 1),
                caps=agent.get("caps"),
                cyclic_processing_limit=agent.get("cyclic_processing_limit"),
                depth_processing_limit=agent.get("depth_processing_limit"),
                accepted_agents=agent.get("accepted_agents"),
                in_selectors=agent.get("in_selectors", []),
            )

            agent_settings.append(agent_def)

        name = agent_group_def.get("name")
        description = agent_group_def.get(
            "description", f"""Agent group : {','.join(agents_names)}"""
        )
        return cls(agent_settings, name, description)

    @classmethod
    def from_bus_message(cls, request):
        """Construct AgentGroupDefinition from a message received in the NATs.

        Args:
            request : The received message.
        """
        agent_settings = []
        agents_names = []
        for agent in request.agents:
            agents_names.append(agent.key)
            replicas = _process_agent_replicas(agent.replicas)
            agent_args = []
            for arg in agent.args:
                agent_arg = defintions.Arg.build(
                    name=arg.name,
                    type=arg.type,
                    value=arg.value,
                )
                agent_args.append(agent_arg)

            agent_def = AgentSettings(
                key=agent.key,
                version=agent.version,
                args=agent_args,
                replicas=replicas,
                caps=agent.caps,
                cyclic_processing_limit=agent.cyclic_processing_limit,
                depth_processing_limit=agent.depth_processing_limit,
                accepted_agents=agent.accepted_agents,
                in_selectors=agent.in_selectors,
            )

            agent_settings.append(agent_def)

        name = request.key.split("/")[-1]
        description = f"Agent group {name}: {','.join(agents_names)}"
        return cls(agent_settings, name, description)


@dataclasses.dataclass
class AssetsDefinition:
    targets: List[base_asset.Asset]
    name: Optional[str] = None
    description: Optional[str] = None

    @classmethod
    def from_yaml(cls, group: io.FileIO):
        """Construct TargetGroupDefinition from yaml file.

        Args:
            group : target group .yaml file.
        """
        target_group_def = loader.load_target_group_yaml(group)
        assets = target_group_def["assets"]
        android_store_assets = assets.get("androidStore", [])
        android_aab_file_assets = assets.get("androidAabFile", [])
        android_apk_file_assets = assets.get("androidApkFile", [])
        ios_store_assets = assets.get("iosStore", [])
        ios_file_assets = assets.get("iosFile", [])
        ip_assets = assets.get("ip", [])
        domain_assets = assets.get("domain", [])
        link_assets = assets.get("link", [])

        assets_def: List[assets.Asset] = []

        for asset in android_aab_file_assets:
            path = asset.get("path")
            url = asset.get("url")
            content = None
            if path is not None:
                content = _load_asset_from_file(asset.get("path", ""))
            if content is None and url is None:
                continue
            assets_def.append(
                android_aab_asset.AndroidAab(
                    content=content, path=path, content_url=url
                )
            )

        for asset in android_apk_file_assets:
            path = asset.get("path")
            url = asset.get("url")
            content = None
            if path is not None:
                content = _load_asset_from_file(asset.get("path", ""))
            if content is None and url is None:
                continue
            assets_def.append(
                android_apk_asset.AndroidApk(
                    content=content, path=path, content_url=url
                )
            )

        for asset in ios_file_assets:
            path = asset.get("path")
            url = asset.get("url")
            content = None
            if path is not None:
                content = _load_asset_from_file(asset.get("path", ""))
            if content is None and url is None:
                continue
            assets_def.append(
                ios_ipa_asset.IOSIpa(content=content, path=path, content_url=url)
            )

        for asset in android_store_assets:
            assets_def.append(
                android_store_asset.AndroidStore(package_name=asset.get("package_name"))
            )

        for asset in ios_store_assets:
            assets_def.append(
                ios_store_asset.IOSStore(bundle_id=asset.get("bundle_id"))
            )

        for asset in ip_assets:
            ip_asset = _parse_ip_asset(asset)
            if ip_asset is not None:
                assets_def.append(ip_asset)

        for asset in domain_assets:
            assets_def.append(domain_name_asset.DomainName(name=asset.get("name")))

        for asset in link_assets:
            assets_def.append(
                link_asset.Link(url=asset.get("url"), method=asset.get("method"))
            )

        return cls(
            targets=assets_def,
            name=target_group_def.get("name"),
            description=target_group_def.get("description"),
        )


def _parse_ip_asset(ip_asset: Dict[str, Any]) -> Optional[base_asset.Asset]:
    ip_string = ip_asset.get("host")
    try:
        ip = ipaddress.ip_address(ip_string)
    except ValueError:
        logger.info(f"Invalid ip address: {ip_string}")
        return None

    if ip.version == 4:
        return ipv4_asset.IPv4(host=ip_string, mask=ip_asset.get("mask"))
    if ip.version == 6:
        return ipv6_asset.IPv6(host=ip_string, mask=ip_asset.get("mask"))
    return None


def _load_asset_from_file(path: str) -> Optional[bytes]:
    path = pathlib.Path(path)
    try:
        content = path.read_bytes()
    except OSError as e:
        logger.error(f"Could not open {path}: {e}.")
        return None
    return content



"""Add agent group asset type mapping

Revision ID: 3aa6ae380275
Revises: 5629037e8282
Create Date: 2024-06-13 14:03:10.208862

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "3aa6ae380275"
down_revision = "5629037e8282"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "asset_type",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("type", sa.String(length=255), nullable=True),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_asset_type")),
    )
    op.create_table(
        "agent_group_asset_type",
        sa.Column("agent_group_id", sa.Integer(), nullable=False),
        sa.Column("asset_type_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["agent_group_id"],
            ["agent_group.id"],
            name=op.f("fk_agent_group_asset_type_agent_group_id_agent_group"),
        ),
        sa.ForeignKeyConstraint(
            ["asset_type_id"],
            ["asset_type.id"],
            name=op.f("fk_agent_group_asset_type_asset_type_id_asset_type"),
        ),
        sa.PrimaryKeyConstraint(
            "agent_group_id", "asset_type_id", name=op.f("pk_agent_group_asset_type")
        ),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("agent_group_asset_type")
    op.drop_table("asset_type")
    # ### end Alembic commands ###



"""mock agent implements the required methods to test the agent's behavior without using external components."""

import dataclasses

import pytest

from typing import List

from ostorlab.agent.message import message as msg


@dataclasses.dataclass
class RawMessage:
    """Raw message as key selector without transformation and message body."""

    key: str
    message: bytes


@dataclasses.dataclass
class AgentRunInstance:
    """An instance run to collect all aspects of an agent instance."""

    emitted_messages: List[msg.Message] = dataclasses.field(default_factory=lambda: [])
    control_messages: List[msg.Message] = dataclasses.field(default_factory=lambda: [])
    raw_messages: List[RawMessage] = dataclasses.field(default_factory=lambda: [])


@pytest.fixture(scope="function")
def agent_mock(mocker) -> List[object]:
    """Fixture patches all the Agent components and returns the list of messages emitted"""
    emitted_messages: List[msg.Message] = []
    mocker.patch(
        "ostorlab.agent.mixins.agent_mq_mixin.AgentMQMixin.mq_init", return_value=None
    )
    mocker.patch(
        "ostorlab.agent.mixins.agent_mq_mixin.AgentMQMixin.mq_run", return_value=None
    )
    mocker.patch(
        "ostorlab.agent.mixins.agent_healthcheck_mixin.AgentHealthcheckMixin.start_healthcheck",
        return_value=None,
    )
    mocker.patch(
        "ostorlab.agent.mixins.agent_healthcheck_mixin.AgentHealthcheckMixin.add_healthcheck",
        return_value=None,
    )
    mocker.patch(
        "ostorlab.agent.mixins.agent_healthcheck_mixin.AgentHealthcheckMixin.__init__",
        return_value=None,
    )

    def mq_send_message(key, message):
        # we need to remove the last part of the key f'{selector}.{uuid.uuid1()}'
        control_message = msg.Message.from_raw("v3.control", message)
        agent_message = msg.Message.from_raw(
            ".".join(key.split(".")[:-1]), control_message.data["message"]
        )
        emitted_messages.append(agent_message)

    mocker.patch(
        "ostorlab.agent.mixins.agent_mq_mixin.AgentMQMixin.mq_send_message",
        side_effect=mq_send_message,
    )
    yield emitted_messages
    emitted_messages = []


@pytest.fixture(scope="function")
def agent_persist_mock(mocker):
    """This fixture patches the Agent persist component and returns the list storage state"""
    storage = {}

    def _set_is_member(key, value):
        """Check values are present in the storage dict."""
        return key in storage and value in storage[key]

    def _set_add(key, *value):
        """Add members to the storage dict and emulate return value."""
        if all(_set_is_member(key, v) for v in value):
            return False
        else:
            for v in value:
                storage.setdefault(key, set()).add(v)
            return True

    def _set_len(key):
        if key in storage:
            return len(storage[key])

    def _set_members(key):
        if key in storage:
            return storage[key]

    def _get(key):
        if key in storage:
            return storage[key]

    def _add(key, value):
        """Check values are present in the storage dict."""
        storage[key] = str(value).encode()

    def _hash_add(hash_name, mapping):
        for k, v in mapping.items():
            if isinstance(v, bytes) is False:
                mapping[k] = str(v).encode()
        storage.setdefault(hash_name, {}).update(mapping)

    def _hash_exists(hash_name, key):
        if isinstance(storage.get(hash_name), dict):
            return key in storage.get(hash_name)
        return False

    def _hash_get(hash_name, key):
        if isinstance(storage.get(hash_name), dict):
            return storage.get(hash_name).get(key, None)
        else:
            return None

    def _hash_get_all(hash_name):
        return storage.get(hash_name, {})

    def _delete(key):
        storage.pop(key)

    def _value_type(key):
        if isinstance(storage.get(key), set):
            return "set"
        elif isinstance(storage.get(key), bytes):
            return "string"
        elif isinstance(storage.get(key), dict):
            return "hash"
        else:
            return "none"

    mocker.patch(
        "ostorlab.agent.mixins.agent_persist_mixin.AgentPersistMixin.set_add",
        side_effect=_set_add,
    )
    mocker.patch(
        "ostorlab.agent.mixins.agent_persist_mixin.AgentPersistMixin.set_is_member",
        side_effect=_set_is_member,
    )
    mocker.patch(
        "ostorlab.agent.mixins.agent_persist_mixin.AgentPersistMixin.set_len",
        side_effect=_set_len,
    )
    mocker.patch(
        "ostorlab.agent.mixins.agent_persist_mixin.AgentPersistMixin.set_members",
        side_effect=_set_members,
    )
    mocker.patch(
        "ostorlab.agent.mixins.agent_persist_mixin.AgentPersistMixin.get",
        side_effect=_get,
    )
    mocker.patch(
        "ostorlab.agent.mixins.agent_persist_mixin.AgentPersistMixin.add",
        side_effect=_add,
    )
    mocker.patch(
        "ostorlab.agent.mixins.agent_persist_mixin.AgentPersistMixin.hash_add",
        side_effect=_hash_add,
    )
    mocker.patch(
        "ostorlab.agent.mixins.agent_persist_mixin.AgentPersistMixin.hash_exists",
        side_effect=_hash_exists,
    )
    mocker.patch(
        "ostorlab.agent.mixins.agent_persist_mixin.AgentPersistMixin.hash_get",
        side_effect=_hash_get,
    )
    mocker.patch(
        "ostorlab.agent.mixins.agent_persist_mixin.AgentPersistMixin.hash_get_all",
        side_effect=_hash_get_all,
    )
    mocker.patch(
        "ostorlab.agent.mixins.agent_persist_mixin.AgentPersistMixin.delete",
        side_effect=_delete,
    )
    mocker.patch(
        "ostorlab.agent.mixins.agent_persist_mixin.AgentPersistMixin.value_type",
        side_effect=_value_type,
    )

    yield storage
    storage = {}


@pytest.fixture(scope="function")
def agent_run_mock(mocker) -> AgentRunInstance:
    """Improved fixture implementation to capture all aspects of an agent run in an `AgentRunInstance` object."""

    agent_run_instance = AgentRunInstance(raw_messages=[], emitted_messages=[])

    mocker.patch(
        "ostorlab.agent.mixins.agent_mq_mixin.AgentMQMixin.mq_init", return_value=None
    )
    mocker.patch(
        "ostorlab.agent.mixins.agent_mq_mixin.AgentMQMixin.mq_run", return_value=None
    )
    mocker.patch(
        "ostorlab.agent.mixins.agent_healthcheck_mixin.AgentHealthcheckMixin.start_healthcheck",
        return_value=None,
    )
    mocker.patch(
        "ostorlab.agent.mixins.agent_healthcheck_mixin.AgentHealthcheckMixin.add_healthcheck",
        return_value=None,
    )
    mocker.patch(
        "ostorlab.agent.mixins.agent_healthcheck_mixin.AgentHealthcheckMixin.__init__",
        return_value=None,
    )

    def mq_send_message(key, message):
        # we need to remove the last part of the key f'{selector}.{uuid.uuid1()}'
        # Control Message.
        control_message = msg.Message.from_raw("v3.control", message)
        agent_run_instance.control_messages.append(control_message)
        # Raw Message.
        agent_run_instance.raw_messages.append(
            RawMessage(key=key, message=control_message.data["message"])
        )
        # Data Message.
        agent_message = msg.Message.from_raw(
            ".".join(key.split(".")[:-1]), control_message.data["message"]
        )
        agent_run_instance.emitted_messages.append(agent_message)

    mocker.patch(
        "ostorlab.agent.mixins.agent_mq_mixin.AgentMQMixin.mq_send_message",
        side_effect=mq_send_message,
    )
    yield agent_run_instance
    agent_run_instance = AgentRunInstance(raw_messages=[], emitted_messages=[])



"""Vulnz List command."""

import logging
from typing import Optional

import click

from ostorlab.cli import console as cli_console
from ostorlab.cli.vulnz import vulnz

console = cli_console.Console()

logger = logging.getLogger(__name__)


@vulnz.command(name="list")
@click.option("--scan-id", "-s", "scan_id", help="Id of the scan.", required=True)
@click.option(
    "--risk-rating",
    "-r",
    "risk_rating",
    help="Filter vulnerabilities by risk ratings. Accept comma-separated ratings.",
)
@click.option("--search", "-sh", help="Search in all content of the vulnerabilities.")
@click.option(
    "--order-by",
    "-o",
    "order_by",
    type=click.Choice(["risk_rating", "title", "id"], case_sensitive=False),
    default="risk_rating",
    help="Order vulnerabilities by specified field. Defaults to risk_rating.",
)
@click.pass_context
def list_cli(
    ctx: click.core.Context,
    scan_id: int,
    risk_rating: Optional[str],
    search: Optional[str],
    order_by: str,
) -> None:
    """CLI command to list vulnerabilities for a scan."""
    runtime_instance = ctx.obj["runtime"]
    console.info(f"Fetching vulnerabilities for scan {scan_id}")
    runtime_instance.list_vulnz(
        scan_id=scan_id,
        filter_risk_rating=risk_rating.strip().split(",") if risk_rating else None,
        search=search,
        order_by=order_by,
    )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ostorlab/agent/message/proto/v3/asset/file/api_schema/wsdl/wsdl.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\nEostorlab/agent/message/proto/v3/asset/file/api_schema/wsdl/wsdl.proto\x12:ostorlab.agent.message.proto.v3.asset.file.api_schema.wsdl\"\'\n\x0f\x41ndroidMetadata\x12\x14\n\x0cpackage_name\x18\x01 \x01(\t\" \n\x0bIOSMetadata\x12\x11\n\tbundle_id\x18\x01 \x01(\t\"%\n\x06Header\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"\xa5\x03\n\x07Message\x12\x0f\n\x07\x63ontent\x18\x01 \x01(\x0c\x12\x0c\n\x04path\x18\x02 \x01(\t\x12\x13\n\x0b\x63ontent_url\x18\x03 \x01(\t\x12g\n\x10\x61ndroid_metadata\x18\x04 \x01(\x0b\x32K.ostorlab.agent.message.proto.v3.asset.file.api_schema.wsdl.AndroidMetadataH\x00\x12_\n\x0cios_metadata\x18\x05 \x01(\x0b\x32G.ostorlab.agent.message.proto.v3.asset.file.api_schema.wsdl.IOSMetadataH\x00\x12\x14\n\x0c\x65ndpoint_url\x18\x06 \x01(\t\x12\x19\n\x0bschema_type\x18\x07 \x01(\t:\x04wsdl\x12Y\n\rextra_headers\x18\x08 \x03(\x0b\x32\x42.ostorlab.agent.message.proto.v3.asset.file.api_schema.wsdl.HeaderB\x10\n\x0emetadata_oneof')



_ANDROIDMETADATA = DESCRIPTOR.message_types_by_name['AndroidMetadata']
_IOSMETADATA = DESCRIPTOR.message_types_by_name['IOSMetadata']
_HEADER = DESCRIPTOR.message_types_by_name['Header']
_MESSAGE = DESCRIPTOR.message_types_by_name['Message']
AndroidMetadata = _reflection.GeneratedProtocolMessageType('AndroidMetadata', (_message.Message,), {
  'DESCRIPTOR' : _ANDROIDMETADATA,
  '__module__' : 'ostorlab.agent.message.proto.v3.asset.file.api_schema.wsdl.wsdl_pb2'
  # @@protoc_insertion_point(class_scope:ostorlab.agent.message.proto.v3.asset.file.api_schema.wsdl.AndroidMetadata)
  })
_sym_db.RegisterMessage(AndroidMetadata)

IOSMetadata = _reflection.GeneratedProtocolMessageType('IOSMetadata', (_message.Message,), {
  'DESCRIPTOR' : _IOSMETADATA,
  '__module__' : 'ostorlab.agent.message.proto.v3.asset.file.api_schema.wsdl.wsdl_pb2'
  # @@protoc_insertion_point(class_scope:ostorlab.agent.message.proto.v3.asset.file.api_schema.wsdl.IOSMetadata)
  })
_sym_db.RegisterMessage(IOSMetadata)

Header = _reflection.GeneratedProtocolMessageType('Header', (_message.Message,), {
  'DESCRIPTOR' : _HEADER,
  '__module__' : 'ostorlab.agent.message.proto.v3.asset.file.api_schema.wsdl.wsdl_pb2'
  # @@protoc_insertion_point(class_scope:ostorlab.agent.message.proto.v3.asset.file.api_schema.wsdl.Header)
  })
_sym_db.RegisterMessage(Header)

Message = _reflection.GeneratedProtocolMessageType('Message', (_message.Message,), {
  'DESCRIPTOR' : _MESSAGE,
  '__module__' : 'ostorlab.agent.message.proto.v3.asset.file.api_schema.wsdl.wsdl_pb2'
  # @@protoc_insertion_point(class_scope:ostorlab.agent.message.proto.v3.asset.file.api_schema.wsdl.Message)
  })
_sym_db.RegisterMessage(Message)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _ANDROIDMETADATA._serialized_start=133
  _ANDROIDMETADATA._serialized_end=172
  _IOSMETADATA._serialized_start=174
  _IOSMETADATA._serialized_end=206
  _HEADER._serialized_start=208
  _HEADER._serialized_end=245
  _MESSAGE._serialized_start=248
  _MESSAGE._serialized_end=669
# @@protoc_insertion_point(module_scope)



"""Handles the public API calls.

Typical usage example:
public_runner = PublicAPIRunner()
public_runner.execute()
"""

from typing import Dict, Any, Optional

import httpx

from ostorlab.apis import request as api_request
from ostorlab.apis.runners import runner

PUBLIC_GRAPHQL_ENDPOINT = "https://api.ostorlab.co/apis/public_graphql"


class PublicAPIRunner(runner.APIRunner):
    """Responsible for the public API calls, and preparing the responses."""

    @property
    def endpoint(self) -> str:
        """API endpoint."""
        return PUBLIC_GRAPHQL_ENDPOINT

    def execute(self, request: api_request.APIRequest) -> Dict[str, Any]:
        """Executes a request using the Public GraphQL API.

        Args:
            request: The request to be executed

        Raises:
            ResponseError: When the API returns an error

        Returns:
            The API response
        """
        response = self._sent_request(request)
        if response.status_code != 200:
            raise runner.ResponseError(
                f'Response status code is {response.status_code}: {response.content.decode(errors="ignore")}'
            )
        data: Dict[str, Any] = response.json()
        errors = data.get("errors")
        if errors is not None and isinstance(errors, list):
            error = errors[0].get("message")
            raise runner.ResponseError(f"Response errors: {error}")
        else:
            return data

    def _sent_request(
        self, request: api_request.APIRequest, headers: Optional[Dict[str, str]] = None
    ) -> httpx.Response:
        """Sends an API request."""
        with httpx.Client(proxy=self._proxy, verify=self._verify) as client:
            return client.post(
                self.endpoint,
                data=request.data,
                timeout=runner.REQUEST_TIMEOUT,
            )



"""Oxo schema."""

import graphene

from ostorlab.serve_app import oxo


class Mutations(oxo.Mutations, graphene.ObjectType):
    pass


class Query(oxo.Query, graphene.ObjectType):
    pass


schema = graphene.Schema(query=Query, mutation=Mutations)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: runtimes/proto/agent_instance_settings.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n,runtimes/proto/agent_instance_settings.proto\x12\x17ostorlab.runtimes.proto\"0\n\x03\x41rg\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0c\n\x04type\x18\x02 \x01(\t\x12\r\n\x05value\x18\x03 \x01(\x0c\"<\n\x0bPortMapping\x12\x13\n\x0bsource_port\x18\x01 \x01(\r\x12\x18\n\x10\x64\x65stination_port\x18\x02 \x01(\r\"\xac\x04\n\x15\x41gentInstanceSettings\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x0f\n\x07\x62us_url\x18\x02 \x01(\t\x12\x1a\n\x12\x62us_exchange_topic\x18\x03 \x01(\t\x12\x1a\n\x12\x62us_management_url\x18\x04 \x01(\t\x12\x11\n\tbus_vhost\x18\x05 \x01(\t\x12*\n\x04\x61rgs\x18\x06 \x03(\x0b\x32\x1c.ostorlab.runtimes.proto.Arg\x12\x13\n\x0b\x63onstraints\x18\x07 \x03(\t\x12\x0e\n\x06mounts\x18\x08 \x03(\t\x12\x16\n\x0erestart_policy\x18\t \x01(\t\x12\x11\n\tmem_limit\x18\n \x01(\x04\x12\x38\n\nopen_ports\x18\x0b \x03(\x0b\x32$.ostorlab.runtimes.proto.PortMapping\x12\x10\n\x08replicas\x18\x0c \x01(\r\x12\x18\n\x10healthcheck_host\x18\r \x01(\t\x12\x18\n\x10healthcheck_port\x18\x0e \x01(\r\x12\x11\n\tredis_url\x18\x0f \x01(\t\x12\x1d\n\x15tracing_collector_url\x18\x10 \x01(\t\x12\x0c\n\x04\x63\x61ps\x18\x11 \x03(\t\x12\x1f\n\x17\x63yclic_processing_limit\x18\x12 \x01(\r\x12\x1e\n\x16\x64\x65pth_processing_limit\x18\x13 \x01(\r\x12\x17\n\x0f\x61\x63\x63\x65pted_agents\x18\x14 \x03(\t\x12\x14\n\x0cin_selectors\x18\x15 \x03(\t')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'runtimes.proto.agent_instance_settings_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _ARG._serialized_start=73
  _ARG._serialized_end=121
  _PORTMAPPING._serialized_start=123
  _PORTMAPPING._serialized_end=183
  _AGENTINSTANCESETTINGS._serialized_start=186
  _AGENTINSTANCESETTINGS._serialized_end=742
# @@protoc_insertion_point(module_scope)



"""Graphene types for the local runtime."""

import collections
import enum
import io
from typing import Optional, List

import graphene
import graphene_sqlalchemy
from graphql.execution import base as graphql_base
from graphene_file_upload import scalars
import ruamel

from ostorlab.runtimes.local.models import models
from ostorlab.serve_app import common
from ostorlab.utils import risk_rating as utils_rik_rating

DEFAULT_NUMBER_ELEMENTS = 15
RISK_RATINGS_ORDER = {
    common.RiskRatingEnum.CRITICAL.name: 8,
    common.RiskRatingEnum.HIGH.name: 7,
    common.RiskRatingEnum.MEDIUM.name: 6,
    common.RiskRatingEnum.LOW.name: 5,
    common.RiskRatingEnum.POTENTIALLY.name: 4,
    common.RiskRatingEnum.HARDENING.name: 3,
    common.RiskRatingEnum.SECURE.name: 2,
    common.RiskRatingEnum.IMPORTANT.name: 1,
    common.RiskRatingEnum.INFO.name: 0,
}
YAML_WIDTH = 100000000


class OxoScanOrderByEnum(graphene.Enum):
    """Enum for the elements to order a scan by."""

    ScanId = enum.auto()
    Title = enum.auto()
    CreatedTime = enum.auto()
    Progress = enum.auto()


OxoAssetTypeEnum = graphene.Enum.from_enum(models.AssetTypeEnum)


class AgentGroupOrderByEnum(graphene.Enum):
    AgentGroupId = enum.auto()
    Name = enum.auto()
    CreatedTime = enum.auto()


class OxoReferenceType(graphene.ObjectType):
    """Graphene object type for a reference."""

    title = graphene.String()
    url = graphene.String()


class OxoKnowledgeBaseVulnerabilityType(graphene.ObjectType):
    """SQLAlchemy object type for a knowledge base vulnerability."""

    references = graphene.List(OxoReferenceType)
    title = graphene.String()
    short_description = graphene.String()
    description = graphene.String()
    recommendation = graphene.String()


OxoRiskRatingEnum = graphene.Enum(
    "OxoRiskRating",
    [(risk.name.upper(), i) for i, risk in enumerate(utils_rik_rating.RiskRating)],
)


class OxoVulnerabilityType(graphene_sqlalchemy.SQLAlchemyObjectType):
    """SQLAlchemy object type for a vulnerability."""

    detail = graphene.Field(OxoKnowledgeBaseVulnerabilityType, required=False)
    cvss_v3_base_score = graphene.Float(required=False)
    risk_rating = graphene.Field(OxoRiskRatingEnum, required=False)

    class Meta:
        """Meta class for the vulnerability object type."""

        model = models.Vulnerability
        only_fields = (
            "id",
            "technical_detail",
            "risk_rating",
            "cvss_v3_vector",
            "dna",
        )

    def resolve_cvss_v3_base_score(
        self: models.Vulnerability, info: graphql_base.ResolveInfo
    ) -> float:
        """Resolve CVSS v3 base score query.

        Args:
            self (models.Vulnerability): The vulnerability object.
            info (graphql_base.ResolveInfo): GraphQL resolve info.

        Returns:
            float: The CVSS v3 base score.
        """
        return common.compute_cvss_v3_base_score(self.cvss_v3_vector)

    def resolve_detail(
        self: models.Vulnerability, info: graphql_base.ResolveInfo
    ) -> OxoKnowledgeBaseVulnerabilityType:
        """Resolve detail query.

        Args:
            self (models.Vulnerability): The vulnerability object.
            info (graphql_base.ResolveInfo): GraphQL resolve info.

        Returns:
            OxoKnowledgeBaseVulnerabilityType: The knowledge base vulnerability.
        """
        with models.Database() as session:
            references = (
                session.query(models.Reference)
                .filter(models.Reference.vulnerability_id == self.id)
                .all()
            )

            return OxoKnowledgeBaseVulnerabilityType(
                title=self.title,
                short_description=self.short_description,
                description=self.description,
                recommendation=self.recommendation,
                references=[
                    OxoReferenceType(title=ref.title, url=ref.url) for ref in references
                ],
            )

    def resolve_risk_rating(
        self: models.Vulnerability, info: graphql_base.ResolveInfo
    ) -> Optional[OxoRiskRatingEnum]:
        """Resolve risk rating of vulnerability"""
        try:
            return OxoRiskRatingEnum[self.risk_rating.name]
        except KeyError:
            return None


class OxoVulnerabilitiesType(graphene.ObjectType):
    """Graphene object type for a list of vulnerabilities."""

    vulnerabilities = graphene.List(OxoVulnerabilityType, required=True)
    page_info = graphene.Field(common.PageInfo, required=False)


class OxoAggregatedKnowledgeBaseVulnerabilityType(graphene.ObjectType):
    """Graphene object type for an aggregated knowledge base vulnerability."""

    highest_risk_rating = graphene.Field(OxoRiskRatingEnum)
    highest_cvss_v3_vector = graphene.String()
    highest_cvss_v3_base_score = graphene.Float()
    kb = graphene.Field(OxoKnowledgeBaseVulnerabilityType)
    vulnerabilities = graphene.Field(
        OxoVulnerabilitiesType,
        detail_titles=graphene.List(graphene.String),
        page=graphene.Int(required=False),
        number_elements=graphene.Int(required=False),
        description="List of vulnerabilities.",
    )

    def resolve_highest_risk_rating(self, info) -> Optional[OxoRiskRatingEnum]:
        try:
            return OxoRiskRatingEnum[self.highest_risk_rating.name]
        except KeyError:
            return None

    def resolve_vulnerabilities(
        self: models.Scan,
        info: graphql_base.ResolveInfo,
        detail_titles: Optional[List[str]] = None,
        page: Optional[int] = None,
        number_elements: int = DEFAULT_NUMBER_ELEMENTS,
    ) -> OxoVulnerabilitiesType:
        """Resolve vulnerabilities query.

        Args:
            self: The scan object.
            info: GraphQL resolve info.
            detail_titles: List of detail titles. Defaults to None.
            page: Page number. Defaults to None.
            number_elements: Number of elements. Defaults to DEFAULT_NUMBER_ELEMENTS.

        Returns:
            OxoVulnerabilitiesType: List of vulnerabilities.
        """
        if number_elements <= 0:
            return OxoVulnerabilitiesType(vulnerabilities=[])

        vulnerabilities = self.vulnerabilities.vulnerabilities.order_by(
            models.Vulnerability.id
        )

        if detail_titles is not None and len(detail_titles) > 0:
            vulnerabilities = vulnerabilities.filter(
                models.Vulnerability.title.in_(detail_titles)
            )

        if page is not None and number_elements > 0:
            p = common.Paginator(vulnerabilities, number_elements)
            page = p.get_page(page)
            page_info = common.PageInfo(
                count=p.count,
                num_pages=p.num_pages,
                has_next=page.has_next(),
                has_previous=page.has_previous(),
            )
            return OxoVulnerabilitiesType(vulnerabilities=page, page_info=page_info)
        else:
            return OxoVulnerabilitiesType(vulnerabilities=vulnerabilities)


class OxoAndroidStoreAssetType(graphene_sqlalchemy.SQLAlchemyObjectType):
    class Meta:
        model = models.AndroidStore
        only_fields = ("id", "package_name", "application_name")

    def resolve_package_name(self, info: graphql_base.ResolveInfo) -> str:
        with models.Database() as session:
            return session.query(models.AndroidStore).get(self.id).package_name

    def resolve_application_name(self, info: graphql_base.ResolveInfo) -> str:
        with models.Database() as session:
            return session.query(models.AndroidStore).get(self.id).application_name


class OxoAndroidStoreAssetInputType(graphene.InputObjectType):
    package_name = graphene.String()
    application_name = graphene.String()


class OxoIOSStoreAssetType(graphene_sqlalchemy.SQLAlchemyObjectType):
    class Meta:
        model = models.IosStore
        only_fields = ("id", "bundle_id", "application_name")

    def resolve_bundle_id(self, info: graphql_base.ResolveInfo) -> str:
        with models.Database() as session:
            return session.query(models.IosStore).get(self.id).bundle_id

    def resolve_application_name(self, info: graphql_base.ResolveInfo) -> str:
        with models.Database() as session:
            return session.query(models.IosStore).get(self.id).application_name


class OxoIOSStoreAssetInputType(graphene.InputObjectType):
    bundle_id = graphene.String()
    application_name = graphene.String()


class OxoAndroidFileAssetType(graphene_sqlalchemy.SQLAlchemyObjectType):
    class Meta:
        model = models.AndroidFile
        only_fields = ("id", "package_name", "path")

    def resolve_package_name(self, info: graphql_base.ResolveInfo) -> str:
        with models.Database() as session:
            return session.query(models.AndroidFile).get(self.id).package_name

    def resolve_path(self, info: graphql_base.ResolveInfo) -> str:
        with models.Database() as session:
            return session.query(models.AndroidFile).get(self.id).path


class OxoAndroidFileAssetInputType(graphene.InputObjectType):
    file = scalars.Upload()
    package_name = graphene.String()


class OxoIOSFileAssetType(graphene_sqlalchemy.SQLAlchemyObjectType):
    class Meta:
        model = models.IosFile
        only_fields = ("id", "bundle_id", "path")

    def resolve_bundle_id(self, info: graphql_base.ResolveInfo) -> str:
        with models.Database() as session:
            return session.query(models.IosFile).get(self.id).bundle_id

    def resolve_path(self, info: graphql_base.ResolveInfo) -> str:
        with models.Database() as session:
            return session.query(models.IosFile).get(self.id).path


class OxoIOSFileAssetInputType(graphene.InputObjectType):
    file = scalars.Upload()
    bundle_id = graphene.String()


class OxoLinkAssetType(graphene_sqlalchemy.SQLAlchemyObjectType):
    class Meta:
        model = models.Link
        only_fields = ("url", "method")


class OxoUrlsAssetType(graphene_sqlalchemy.SQLAlchemyObjectType):
    links = graphene.List(OxoLinkAssetType, required=False)

    class Meta:
        model = models.Urls
        only_fields = ("id",)

    def resolve_links(self, info) -> List[OxoLinkAssetType]:
        with models.Database() as session:
            links = session.query(models.Link).filter_by(urls_asset_id=self.id).all()
            return [
                OxoLinkAssetType(url=link.url, method=link.method) for link in links
            ]


class OxoIPRangeAssetType(graphene_sqlalchemy.SQLAlchemyObjectType):
    class Meta:
        model = models.IPRange
        only_fields = ("host", "mask")


class OxoNetworkAssetType(graphene_sqlalchemy.SQLAlchemyObjectType):
    networks = graphene.List(OxoIPRangeAssetType, required=False)

    class Meta:
        model = models.Network
        only_fields = ("id",)

    def resolve_networks(self, info) -> List[OxoIPRangeAssetType]:
        with models.Database() as session:
            ips = (
                session.query(models.IPRange).filter_by(network_asset_id=self.id).all()
            )
            return [OxoIPRangeAssetType(host=ip.host, mask=ip.mask) for ip in ips]


class OxoDomainNameAssetType(graphene_sqlalchemy.SQLAlchemyObjectType):
    class Meta:
        model = models.DomainName
        only_fields = "name"


class OxoDomainNameAssetsType(graphene_sqlalchemy.SQLAlchemyObjectType):
    domain_names = graphene.List(OxoDomainNameAssetType, required=False)

    class Meta:
        model = models.DomainAsset
        only_fields = ("id",)

    def resolve_domain_names(
        self, info: graphql_base.ResolveInfo
    ) -> List[OxoDomainNameAssetType]:
        """Resolve domain names query.

        Args:
            self: The domain asset object.
            info: GraphQL resolve info.

        Returns:
            List of domain names.
        """
        with models.Database() as session:
            domain_names = (
                session.query(models.DomainName)
                .filter_by(domain_asset_id=self.id)
                .all()
            )
            return [
                OxoDomainNameAssetType(name=domain_name.name)
                for domain_name in domain_names
            ]


class OxoAssetType(graphene.Union):
    class Meta:
        model = models.Asset
        types = (
            OxoAndroidFileAssetType,
            OxoIOSFileAssetType,
            OxoAndroidStoreAssetType,
            OxoIOSStoreAssetType,
            OxoUrlsAssetType,
            OxoNetworkAssetType,
            OxoDomainNameAssetsType,
        )


class OxoAgentArgumentType(graphene_sqlalchemy.SQLAlchemyObjectType):
    """Graphene object type for a list of agent arguments."""

    value = common.Bytes(required=False)

    class Meta:
        """Meta class for the agent arguments object type."""

        model = models.AgentArgument
        only_fields = (
            "name",
            "type",
            "description",
        )

    def resolve_value(
        self: models.AgentArgument, info: graphql_base.ResolveInfo
    ) -> bytes:
        """Resolve agent argument value query.

        Args:
            self (models.AgentArgument): The agent argument object.
            info (graphql_base.ResolveInfo): GraphQL resolve info.

        Returns:
            common.Bytes: The value of the agent argument.
        """
        return self.value


class OxoAgentArgumentsType(graphene.ObjectType):
    """Graphene object type for a list of agent arguments."""

    args = graphene.List(OxoAgentArgumentType, required=True)


class OxoAgentType(graphene_sqlalchemy.SQLAlchemyObjectType):
    """SQLAlchemy object type for an agent."""

    args = graphene.Field(OxoAgentArgumentsType, required=True)

    class Meta:
        """Meta class for the agent object type."""

        model = models.Agent
        only_fields = ("key",)

    def resolve_args(
        self: models.Agent, info: graphql_base.ResolveInfo
    ) -> OxoAgentArgumentsType:
        """Resolve agent arguments query.

        Args:
            self (models.Agent): The agent object.
            info (graphql_base.ResolveInfo): GraphQL resolve info.

        Returns:
            AgentArgumentsType: List of agent arguments.
        """
        with models.Database() as session:
            args = session.query(models.AgentArgument).filter(
                models.AgentArgument.agent_id == self.id
            )
            return OxoAgentArgumentsType(args=args)


class OxoAgentsType(graphene.ObjectType):
    """Graphene object type for a list of agents."""

    agents = graphene.List(OxoAgentType, required=True)
    page_info = graphene.Field(
        common.PageInfo,
        required=False,
    )


class OxoAgentGroupType(graphene_sqlalchemy.SQLAlchemyObjectType):
    """SQLAlchemy object type for an agent group."""

    key = graphene.String()
    agents = graphene.Field(
        OxoAgentsType,
        required=True,
        page=graphene.Int(required=False),
        number_elements=graphene.Int(required=False),
    )
    asset_types = graphene.List(graphene.String)
    yaml_source = graphene.String()

    class Meta:
        """Meta class for the agent group object type."""

        model = models.AgentGroup

        only_fields = (
            "id",
            "name",
            "description",
            "created_time",
        )

    def resolve_key(self: models.AgentGroup, info: graphql_base.ResolveInfo) -> str:
        """Resolve key query.
        Args:
            self (models.AgentGroup): The agent group object.
            info (graphql_base.ResolveInfo): GraphQL resolve info.
        Returns:
            str: The key of the agent group.
        """
        return (
            f"agentgroup//{self.name}"
            if self.name is not None
            else f"agentgroup//{self.id}"
        )

    def resolve_agents(
        self: models.AgentGroup,
        info: graphql_base.ResolveInfo,
        page: int = None,
        number_elements: int = DEFAULT_NUMBER_ELEMENTS,
    ) -> OxoAgentsType:
        """Resolve agents query.
        Args:
            self (models.AgentGroup): The agent group object.
            info (graphql_base.ResolveInfo): GraphQL resolve info.
        Returns:
            AgentsType: List of agents.
        """
        if number_elements <= 0:
            return OxoAgentsType(agents=[])

        with models.Database() as session:
            agents = (
                session.query(models.AgentGroup)
                .filter(models.AgentGroup.id == self.id)
                .first()
                .agents
            )
            if page is not None and number_elements > 0:
                p = common.Paginator(agents, number_elements)
                page = p.get_page(page)
                page_info = common.PageInfo(
                    count=p.count,
                    num_pages=p.num_pages,
                    has_next=page.has_next(),
                    has_previous=page.has_previous(),
                )
                return OxoAgentsType(agents=page, page_info=page_info)
            else:
                return OxoAgentsType(agents=agents)

    def resolve_asset_types(
        self: models.AgentGroup, info: graphql_base.ResolveInfo
    ) -> List[str]:
        """Resolve asset types query.
        Args:
            self (models.AgentGroup): The agent group object.
            info (graphql_base.ResolveInfo): GraphQL resolve info.
        Returns:
            List[str]: The asset types of the agent group.
        """
        with models.Database() as session:
            asset_types = session.query(models.AgentGroup).get(self.id).asset_types
            return [asset.type.name for asset in asset_types]

    def resolve_yaml_source(
        self: models.AgentGroup, info: graphql_base.ResolveInfo
    ) -> str:
        """Resolve yaml source query.
        Args:
            self: The agent group object.
            info: GraphQL resolve info.
        Returns:
            The yaml source of the agent group.
        """
        yaml = ruamel.yaml.YAML(typ="safe")
        yaml.width = YAML_WIDTH
        agent_group_definition = {"kind": "AgentGroup", "agents": []}

        if self.name is not None:
            agent_group_definition["name"] = self.name

        if self.description is not None:
            agent_group_definition["description"] = self.description

        with models.Database() as session:
            agent_group = session.query(models.AgentGroup).get(self.id)
            agents = agent_group.agents
            for agent in agents:
                agent_definition = {
                    "key": agent.key,
                    "args": [],
                }

                args = (
                    session.query(models.AgentArgument)
                    .filter_by(agent_id=agent.id)
                    .all()
                )
                for arg in args:
                    value = models.AgentArgument.from_bytes(arg.type, arg.value)
                    arg_dict = {
                        "name": arg.name,
                        "type": arg.type,
                        "description": arg.description,
                    }
                    if value is not None:
                        arg_dict["value"] = value
                    agent_definition["args"].append(arg_dict)

                agent_group_definition["agents"].append(agent_definition)

        string_yaml_io = io.StringIO()
        yaml.dump(agent_group_definition, string_yaml_io)
        agent_group_definition_yaml = string_yaml_io.getvalue()
        return agent_group_definition_yaml


class OxoAgentGroupsType(graphene.ObjectType):
    agent_groups = graphene.List(OxoAgentGroupType, required=True)
    page_info = graphene.Field(common.PageInfo, required=False)


class OxoIPRangeInputType(graphene.InputObjectType):
    host = graphene.String(required=True)
    mask = graphene.String(required=False)


class OxoLinkInputType(graphene.InputObjectType):
    url = graphene.String(required=True)
    method = graphene.String(required=False, default_value="GET")


class OxoDomainNameInputType(graphene.InputObjectType):
    name = graphene.String(required=True)


class OxoAssetInputType(graphene.InputObjectType):
    android_apk_file = graphene.List(OxoAndroidFileAssetInputType)
    android_aab_file = graphene.List(OxoAndroidFileAssetInputType)
    ios_file = graphene.List(OxoIOSFileAssetInputType)
    android_store = graphene.List(OxoAndroidStoreAssetInputType)
    ios_store = graphene.List(OxoIOSStoreAssetInputType)
    link = graphene.List(OxoLinkInputType)
    ip = graphene.List(OxoIPRangeInputType)
    domain = graphene.List(OxoDomainNameInputType)


class OxoAgentArgumentInputType(graphene.InputObjectType):
    """Input object type for an agent argument."""

    name = graphene.String(required=True)
    type = graphene.String(required=True)
    description = graphene.String(required=False)
    value = common.Bytes(required=False)


class OxoAgentGroupAgentCreateInputType(graphene.InputObjectType):
    """Input object type for creating an agent group agent."""

    key = graphene.String(required=True)
    args = graphene.List(OxoAgentArgumentInputType, required=False, default_value=[])


class OxoAgentGroupCreateInputType(graphene.InputObjectType):
    """Input object type for creating an agent group."""

    name = graphene.String(required=False)
    description = graphene.String(required=True)
    agents = graphene.List(OxoAgentGroupAgentCreateInputType, required=True)
    asset_types = graphene.List(OxoAssetTypeEnum, required=False, default_value=[])


class OxoAgentScanInputType(graphene.InputObjectType):
    """Input object type for scan"""

    title = graphene.String(required=False)
    asset_ids = graphene.List(graphene.Int, required=True)
    agent_group_id = graphene.Int(required=True)


class OxoScanType(graphene_sqlalchemy.SQLAlchemyObjectType):
    """SQLAlchemy object type for a scan."""

    vulnerabilities = graphene.Field(
        OxoVulnerabilitiesType,
        page=graphene.Int(required=False),
        number_elements=graphene.Int(required=False),
        detail_titles=graphene.List(graphene.String, required=False),
        vuln_ids=graphene.List(graphene.Int, required=False),
        description="List of vulnerabilities.",
    )
    kb_vulnerabilities = graphene.Field(
        graphene.List(OxoAggregatedKnowledgeBaseVulnerabilityType),
        detail_title=graphene.String(required=False),
        description="List of aggregated knowledge base vulnerabilities.",
    )
    message_status = graphene.String()
    progress = graphene.String()
    risk_rating = graphene.String()
    assets = graphene.List(OxoAssetType)
    agent_group = graphene.Field(OxoAgentGroupType)

    class Meta:
        """Meta class for the scan object type."""

        model = models.Scan
        description = "Scan object."
        only_fields = (
            "id",
            "title",
            "created_time",
        )

    def resolve_progress(self: models.Scan, info: graphql_base.ResolveInfo) -> str:
        """Resolve progress query.

        Args:
            self (models.Scan): The scan object.
            info (graphql_base.ResolveInfo): GraphQL resolve info.

        Returns:
            str: The progress of the scan.
        """

        return self.progress.name

    def resolve_risk_rating(
        self: models.Scan, info: graphql_base.ResolveInfo
    ) -> Optional[str]:
        """Resolve risk rating query.
        Args:
            self (models.Scan): The scan object.
            info (graphql_base.ResolveInfo): GraphQL resolve info.
        Returns:
            str: The risk rating of the scan.
        """
        return self.risk_rating.name if self.risk_rating is not None else None

    def resolve_assets(
        self,
        info: graphql_base.ResolveInfo,
    ):
        """Resolve asset query.

        Args:
            self (models.Scan): The scan object.
            info (graphql_base.ResolveInfo): GraphQL resolve info.

        Returns:
            List[OxoAssetType]: The asset of the scan.
        """
        with models.Database() as session:
            assets = session.query(models.Asset).filter_by(scan_id=self.id).all()
            return assets

    def resolve_vulnerabilities(
        self: models.Scan,
        info: graphql_base.ResolveInfo,
        detail_titles: Optional[List[str]] = None,
        vuln_ids: Optional[List[int]] = None,
        page: Optional[int] = None,
        number_elements: int = DEFAULT_NUMBER_ELEMENTS,
    ) -> OxoVulnerabilitiesType:
        """Resolve vulnerabilities query.

        Args:
            self: The scan object.
            info: GraphQL resolve info.
            detail_titles: List of detail titles. Defaults to None.
            vuln_ids: List of vulnerability ids. Defaults to None.
            page: Page number. Defaults to None.
            number_elements: Number of elements. Defaults to DEFAULT_NUMBER_ELEMENTS.

        Returns:
            OxoVulnerabilitiesType: List of vulnerabilities.
        """
        if number_elements <= 0:
            return OxoVulnerabilitiesType(vulnerabilities=[])

        with models.Database() as session:
            vulnerabilities = session.query(models.Vulnerability).filter(
                models.Vulnerability.scan_id == self.id
            )

            if vuln_ids is not None and len(vuln_ids) > 0:
                vulnerabilities = vulnerabilities.filter(
                    models.Vulnerability.id.in_(vuln_ids)
                )

            if detail_titles is not None and len(detail_titles) > 0:
                vulnerabilities = vulnerabilities.filter(
                    models.Vulnerability.title.in_(detail_titles)
                )

            vulnerabilities = vulnerabilities.order_by(models.Vulnerability.id)

            if page is not None and number_elements > 0:
                p = common.Paginator(vulnerabilities, number_elements)
                page = p.get_page(page)
                page_info = common.PageInfo(
                    count=p.count,
                    num_pages=p.num_pages,
                    has_next=page.has_next(),
                    has_previous=page.has_previous(),
                )
                return OxoVulnerabilitiesType(vulnerabilities=page, page_info=page_info)
            else:
                return OxoVulnerabilitiesType(vulnerabilities=vulnerabilities)

    def resolve_kb_vulnerabilities(
        self: models.Scan,
        info: graphql_base.ResolveInfo,
        detail_title: Optional[str] = None,
        page: Optional[int] = None,
        number_elements: int = DEFAULT_NUMBER_ELEMENTS,
    ) -> list[OxoAggregatedKnowledgeBaseVulnerabilityType]:
        """Resolve knowledge base vulnerabilities query.

        Args:
            self: The scan object.
            info: GraphQL resolve info.
            detail_title: The detail title. Defaults to None.

        Returns:
            list[OxoAggregatedKnowledgeBaseVulnerabilityType]: List of aggregated knowledge base vulnerabilities.
        """
        aggregated_kb = OxoScanType._build_kb_vulnerabilities(self, detail_title)
        return aggregated_kb

    def resolve_message_status(
        self: models.Scan, info: graphql_base.ResolveInfo
    ) -> str:
        """Resolve message status query.

        Args:
            self (models.Scan): The scan object.
            info (graphql_base.ResolveInfo): GraphQL resolve info.

        Returns:
            str: The message status of the scan.
        """
        with models.Database() as session:
            scan_statuses = session.query(models.ScanStatus).filter(
                models.ScanStatus.scan_id == self.id
            )
            message_statuses = [
                s.value for s in scan_statuses if s.key == "message_status"
            ]
            if message_statuses is not None and len(message_statuses) > 0:
                return message_statuses[-1]

    def resolve_agent_group(
        self: models.Scan, info: graphql_base.ResolveInfo
    ) -> Optional[OxoAgentGroupType]:
        """Resolve agent group.

        Args:
            self: The scan object.
            info: GraphQL resolve info.

        Returns:
            str: The message status of the scan.
        """
        with models.Database() as session:
            if self.agent_group_id is not None:
                return session.query(models.AgentGroup).get(self.agent_group_id)

    @staticmethod
    def _build_kb_vulnerabilities(
        scan: models.Scan, detail_title: Optional[str] = None
    ) -> list[OxoAggregatedKnowledgeBaseVulnerabilityType]:
        """Build knowledge base vulnerabilities.

        Args:
            scan: The scan object.
            detail_title: The detail title. Defaults to None.

        Returns:
            list[OxoAggregatedKnowledgeBaseVulnerabilityType]: List of aggregated knowledge base vulnerabilities.
        """
        with models.Database() as session:
            vulnerabilities = session.query(models.Vulnerability).filter(
                models.Vulnerability.scan_id == scan.id
            )
            if detail_title is not None:
                vulnerabilities = vulnerabilities.filter(
                    models.Vulnerability.title == detail_title
                )

            kbs = vulnerabilities.group_by(
                models.Vulnerability.title,
                models.Vulnerability.short_description,
                models.Vulnerability.recommendation,
            ).all()

            distinct_vulnz = vulnerabilities.distinct(
                models.Vulnerability.risk_rating,
                models.Vulnerability.cvss_v3_vector,
            ).all()

            kb_dict = collections.defaultdict(list)
            cvss_dict = collections.defaultdict(list)

            aggregated_kb = []
            for vuln in distinct_vulnz:
                kb_dict[vuln.title].append(vuln.risk_rating)
                cvss_dict[vuln.title].append(vuln.cvss_v3_vector)

            for kb in kbs:
                kb_vulnerabilities = vulnerabilities.filter(
                    models.Vulnerability.title == kb.title
                )
                highest_risk_rating = max(
                    kb_dict[kb.title], key=lambda risk: RISK_RATINGS_ORDER[risk.name]
                )
                if cvss_dict is not None:
                    cvss_v3_vectors = [
                        v
                        for v in cvss_dict[kb.title]
                        if v is not None
                        and common.compute_cvss_v3_base_score(v) is not None
                    ]
                if len(cvss_v3_vectors) > 0:
                    highest_cvss_v3_vector = max(
                        cvss_v3_vectors,
                        key=lambda vector: common.compute_cvss_v3_base_score(vector),
                    )
                else:
                    highest_cvss_v3_vector = kb.cvss_v3_vector or None
                references = (
                    session.query(models.Reference)
                    .filter(
                        models.Reference.vulnerability_id
                        == kb_vulnerabilities.first().id
                    )
                    .distinct(models.Reference.title)
                ).all()
                aggregated_kb.append(
                    OxoAggregatedKnowledgeBaseVulnerabilityType(
                        highest_risk_rating=highest_risk_rating,
                        highest_cvss_v3_vector=highest_cvss_v3_vector,
                        highest_cvss_v3_base_score=common.compute_cvss_v3_base_score(
                            highest_cvss_v3_vector
                        ),
                        kb=OxoKnowledgeBaseVulnerabilityType(
                            title=kb.title,
                            short_description=kb.short_description,
                            description=kb.description,
                            recommendation=kb.recommendation,
                            references=[
                                OxoReferenceType(title=ref.title, url=ref.url)
                                for ref in references
                            ],
                        ),
                        vulnerabilities=OxoVulnerabilitiesType(
                            vulnerabilities=kb_vulnerabilities
                        ),
                    )
                )

            return aggregated_kb


class OxoScansType(graphene.ObjectType):
    """Graphene object type for a list of scans."""

    scans = graphene.List(OxoScanType, required=True)
    page_info = graphene.Field(common.PageInfo, required=False)



"""Module Responsible for subscribing to nats for different subjects."""

import logging
import asyncio
import datetime
from typing import List, Optional

import docker
from docker.models import services
from nats.js import errors as jetstream_errors
from nats import errors as nats_errors

from ostorlab.apis import scanner_config
from ostorlab.apis.runners import authenticated_runner
from ostorlab.scanner import handler as scanner_handler
from ostorlab.scanner import scanner_conf
from ostorlab.scanner import callbacks
from ostorlab.utils import scanner_state_reporter

logger = logging.getLogger(__name__)


WAIT_CHECK_MESSAGES = datetime.timedelta(seconds=5)

logger = logging.getLogger(__name__)


class ScanHandler:
    """Class responsible for handling the subscription to bus handler."""

    def __init__(self, state_reporter: scanner_state_reporter.ScannerStateReporter):
        self._bus_handlers = []
        self._state_reporter = state_reporter
        self._docker_client = docker.from_env()

    async def close(self) -> None:
        for handler in self._bus_handlers:
            await handler.close()

    async def _subscribe(
        self,
        config: scanner_conf.ScannerConfig,
        subject,
        queue,
        stream,
        start_at="first",
    ) -> scanner_handler.BusHandler:
        bus_handler = await self._create_bus_handler(config)
        asyncio.create_task(bus_handler.ensure_running_handler())
        await bus_handler.connect()
        await bus_handler.add_stream(name=stream, subjects=[subject])
        await bus_handler.subscribe(
            subject=subject,
            start_at=start_at,
            durable_name=queue,
        )
        self._bus_handlers.append(bus_handler)
        return bus_handler

    async def subscribe_all(self, config: scanner_conf.ScannerConfig) -> None:
        bus_handlers = []
        for bus_conf in config.subject_bus_configs:
            bus_handler = await self._subscribe(
                config=config,
                subject=bus_conf.subject,
                queue=bus_conf.queue,
                start_at="first",
                stream=bus_conf.queue,
            )
            logger.info("subscribing to %s", bus_conf.subject)
            bus_handlers.append(bus_handler)

        for bus_handler in bus_handlers:
            await self.handle_messages(bus_handler, config)

    async def handle_messages(
        self,
        bus_handler: scanner_handler.BusHandler,
        config: scanner_conf.ScannerConfig,
    ) -> None:
        """Scan handler method responsible for fetching messages from the streaming server,
        parse the messages and trigger the scan.
        The message is acknowledged after the scan is created, and the scan handler can't fetch new messages,
        until the current scan is no longer running.

        Args:
            bus_handler: instance for performing BUS operations.
            config: The scanner configuration; holds credentials for the registry & streaming server.
        """
        scan_id = None
        while True:
            if _is_scan_running(self._docker_client, scan_id=scan_id) is True:
                await asyncio.sleep(WAIT_CHECK_MESSAGES.seconds)
            else:
                try:
                    scan_id = await self._handle_message(bus_handler, config)
                except nats_errors.TimeoutError:
                    # No available message to fetch.
                    await asyncio.sleep(WAIT_CHECK_MESSAGES.seconds)

    async def _handle_message(
        self,
        bus_handler: scanner_handler.BusHandler,
        config: scanner_conf.ScannerConfig,
    ) -> str:
        """Fetch, parse a single message and trigger the corresponding scan."""
        async for msg, request in bus_handler.process_message():
            if request is not None and msg is not None:
                try:
                    scan_id = callbacks.start_scan(
                        subject=msg.subject,
                        request=request,
                        state_reporter=self._state_reporter,
                        registry_conf=config.registry_conf,
                    )
                    await msg.ack()
                    return scan_id
                except Exception as e:
                    logger.exception("Exception: %s", e)
                    await msg.nak()

    async def _create_bus_handler(
        self, config: scanner_conf.ScannerConfig
    ) -> scanner_handler.BusHandler:
        bus_handler = scanner_handler.BusHandler(
            bus_url=config.bus_url,
            cluster_id=config.bus_cluster_id,
            name=config.bus_client_name,
        )
        return bus_handler


def _is_scan_running(
    client: docker.DockerClient, scan_id: Optional[str] = None
) -> bool:
    """Returns True, if docker services with `ostorlab.universe` label exist,
    False otherwise.
    """
    if scan_id is None:
        return False
    scan_services: List[services.Service] = client.services.list(
        filters={"label": f"ostorlab.universe={str(scan_id)}"}
    )
    return len(scan_services) > 0


async def connect_nats(
    config: scanner_conf.ScannerConfig,
    scanner_id: str,
    state_reporter: scanner_state_reporter.ScannerStateReporter,
) -> ScanHandler:
    """connecting to nats.

    Args:
        config: The scanner configuration; holds credentials for the registry & streaming server.
        scanner_id: The scanner identifier.
        state_reporter: instance responsible for reporting the scanner state.
    """
    try:
        logger.info("starting bus runner for scanner %s", scanner_id)
        scan_handler = ScanHandler(state_reporter)
        logger.info("connected, subscribing to plans channels ...")
        await scan_handler.subscribe_all(config)
        logger.info("subscribed")
        return scan_handler
    except jetstream_errors.ServiceUnavailableError as e:
        logger.exception("Failed to establish connection to NATs: %s", e)


async def subscribe_nats(
    api_key: str,
    scanner_id: str,
    state_reporter: scanner_state_reporter.ScannerStateReporter,
) -> None:
    """Fetching the scanner configuration and subscribing to nats.

    Args:
        api_key: The key to connect to ostorlab.
        scanner_id: The scanner identifier.
        state_reporter: instance responsible for reporting the scanner state.
    """
    logger.info("Fetching scanner configuration.")
    runner = authenticated_runner.AuthenticatedAPIRunner(api_key=api_key)
    data = runner.execute(scanner_config.ScannerConfigAPIRequest(scanner_id=scanner_id))
    config = scanner_conf.ScannerConfig.from_json(data)

    if config is None:
        logger.error("No config found to start the connection.")
        return

    logger.info("Connecting to nats.")
    await connect_nats(config, scanner_id, state_reporter)



"""common utilities for the flask app."""

import collections
import inspect
import io
import json
import zipfile
from functools import cached_property
from math import ceil
import struct
from typing import Optional, Union

import cvss
import graphene
from graphql.language import ast
from graphene.types import scalars


class PageInfo(graphene.ObjectType):
    """Page info object type."""

    count = graphene.Int()
    num_pages = graphene.Int()
    has_next = graphene.Boolean()
    has_previous = graphene.Boolean()


class SortEnum(graphene.Enum):
    """Sort enum, for the sorting order of the results."""

    Asc = 1
    Desc = 2


class RiskRatingEnum(graphene.Enum):
    """Enum for the risk rating of a scan."""

    CRITICAL = "Critical"
    HIGH = "High"
    MEDIUM = "Medium"
    LOW = "Low"
    POTENTIALLY = "Potentially"
    HARDENING = "Hardening"
    SECURE = "Secure"
    IMPORTANT = "Important"
    INFO = "Info"


class Bytes(scalars.Scalar):
    """
    The `Bytes` scalar type represents binary data in a bytes format.
    """

    @staticmethod
    def coerce_bytes(
        value: Union[str, bytes, memoryview, list, float, int, dict, bool],
    ) -> bytes:
        """Coerce a value to bytes.

        Args:
            value: Value to coerce.

        Returns:
            bytes: Coerced value.
        """
        if isinstance(value, bytes):
            return value
        elif isinstance(value, memoryview):
            return value.tobytes()
        elif isinstance(value, str):
            return Bytes._rawbytes(value)
        elif isinstance(value, list):
            return json.dumps(value).encode(encoding="utf-8")
        elif isinstance(value, float) or isinstance(value, int):
            return struct.pack("d", value)
        elif isinstance(value, dict):
            return json.dumps(value).encode(encoding="utf-8")
        elif isinstance(value, bool):
            return (1 if value is True else 0).to_bytes(1, byteorder="big")
        else:
            raise NotImplementedError(f"Bytes scalar coerce error from {type(value)}")

    serialize = coerce_bytes
    parse_value = coerce_bytes

    @staticmethod
    def parse_literal(asst: ast.Value) -> Optional[int]:
        """Parse a literal value."""
        raise NotImplementedError("ast is not supported")

    @staticmethod
    def _rawbytes(s: str) -> bytes:
        """Convert a string to raw bytes without encoding"""
        outlist = []
        for cp in s:
            num = ord(cp)
            if num <= 255:
                outlist.append(struct.pack("B", num))
            elif num < 65535:
                outlist.append(struct.pack(">H", num))
            else:
                b = (num & 0xFF0000) >> 16
                H = num & 0xFFFF
                outlist.append(struct.pack(">bH", b, H))
        return b"".join(outlist)


def compute_cvss_v3_base_score(vector: Optional[str]) -> Optional[float]:
    """Compute the CVSS v3 base score from the vector.

    Args:
        vector (str | None): CVSS v3 vector.

    Returns:
        CVSS v3 base score or None if the vector is invalid.
    """
    if vector is None:
        return None
    try:
        return cvss.CVSS3(vector).scores()[0]
    except cvss.CVSS3Error:
        return None


class Page(collections.abc.Sequence):
    """A single page in a paginated list of objects."""

    def __init__(self, object_list, number, paginator):
        """
        initialize the page object.
        Args:
            object_list: The list of objects in the page.
            number: page number.
            paginator: paginator object.
        """
        self.object_list = object_list
        self.number = number
        self.paginator = paginator

    def __repr__(self):
        """Return the string representation of the page object."""
        return "<Page %s of %s>" % (self.number, self.paginator.num_pages)

    def __len__(self):
        """Return the number of objects in the page."""
        return len(self.object_list)

    def __getitem__(self, index):
        """Return the object at the given index in the page."""
        if isinstance(index, (int, slice)) is False:
            raise TypeError(
                "Page indices must be integers or slices, not %s."
                % type(index).__name__
            )
        if isinstance(self.object_list, list) is False:
            self.object_list = list(self.object_list)
        return self.object_list[index]

    def has_next(self) -> bool:
        """Return True if there is a next page."""
        return self.number < self.paginator.num_pages

    def has_previous(self) -> bool:
        """Return True if there is a previous page."""
        return self.number > 1


class Paginator:
    """A paginator object for paginating a list of objects."""

    def __init__(self, object_list: list[any], per_page):
        """Initialize the paginator object.
        Args:
            object_list: The list of objects to paginate.
            per_page: The number of objects per page.
        """
        self.object_list = object_list
        self.per_page = per_page

    @cached_property
    def count(self) -> int:
        """Return the total number of objects, across all pages."""
        c = getattr(self.object_list, "count", None)
        if callable(c) is True and inspect.isbuiltin(c) is False:
            return c()
        return len(self.object_list)

    @cached_property
    def num_pages(self) -> int:
        """Return the total number of pages."""
        if self.count == 0:
            return 0
        hits = max(1, self.count)
        return ceil(hits / self.per_page)

    def get_page(self, number: int) -> Page:
        """Return a valid page, even if the page argument isn't a number or isn't in range.
        Args:
            number: The page number.
        Returns: The page object.
        """
        return self.page(number)

    def page(self, number: int) -> Page:
        """Return a Page object for the given 1-based page number
        Args:
            number: The page number.

        Returns: The page object.
        """
        bottom = (number - 1) * self.per_page
        top = bottom + self.per_page
        if top >= self.count:
            top = self.count
        return Page(
            object_list=self.object_list[bottom:top], number=number, paginator=self
        )

    def _get_page(self, *args, **kwargs) -> Page:
        """Return an instance of a single page."""
        return Page(*args, **kwargs)


def is_apk(file_content: bytes) -> bool:
    """Check if a file is an apk.

    Args:
        file_content: File to check if it's an apk or not.

    Returns:
        True if the file is a valid apk file, False otherwise.
    """

    try:
        with zipfile.ZipFile(io.BytesIO(file_content)) as o:
            if "AndroidManifest.xml" in o.namelist():
                return True
            return False
    except zipfile.BadZipFile:
        return False


def is_xapk(file_content: bytes) -> bool:
    """Check if a file is a xapk bundle.

    Args:
        file_content: File to check if it's xapk application or not.

    Returns:
        True if the file is a valid xapk file, False otherwise.
    """

    try:
        with zipfile.ZipFile(io.BytesIO(file_content)) as o:
            return all(file_name.endswith(".apk") for file_name in o.namelist())
    except zipfile.BadZipFile:
        return False


def is_aab(file_content: bytes) -> bool:
    """Check if file is an AAB file.

    Args:
        file_content: File to check if it's an aab or not.

    Returns:
        True if the file is a valid aab file, False otherwise.
    """

    try:
        with zipfile.ZipFile(io.BytesIO(file_content)) as o:
            if "BundleConfig.pb" in o.namelist():
                return True
            return False
    except zipfile.BadZipFile:
        return False



"""Module responsible for interacting with the OpenAI API to generate KB entries."""

import dataclasses
import enum
import json
import logging
import os
import pathlib
import re
from typing import Any

import click
import openai
import tenacity
from openai import openai_object
from openai.api_resources import chat_completion

DEMO_LANGS = ["Flutter", "Swift", "Kotlin"]

logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s"
)

OPENAI_API_KEY = os.getenv("OPENAI_API_KEY")
MODEL_NAME = "gpt-3.5-turbo"

DESCRIPTION_TEMPLATE = """
# %%VULNERABILITY_NAME%%

%%VULNERABILITY_DESCRIPTION%%

### Examples

#### Dart

```dart
%%FLUTTER_CODE%%
```

#### Swift

```swift
%%SWIFT_CODE%%
```

#### Kotlin

```kotlin
%%KOTLIN_CODE%%
```
"""

RECOMMENDATION_TEMPLATE = """
# %%VULNERABILITY_NAME%%

%%RECOMMENDATION%%

# Code Examples:

### Dart

```dart
%%FLUTTER_CODE%%
```

### Swift

```swift
%%SWIFT_CODE%%
```

### Kotlin

```kotlin
%%KOTLIN_CODE%%
```
"""

META_TEMPLATE = """
{
   "risk_rating":"[hardening/info/low/medium/high]",
   "short_description":"[short description of the vulnerability]",
   "references":{
      "Vulnerability (SOURCE)":"[SOURCE LINK]",
      ...
   },
   "title":"[Vulnerability title]",
   "privacy_issue":[true/false],
   "security_issue":[true/false],
   "categories":{
       "OWASP_MASVS_L1":[
         "MSTG reference" 
      ],
      "OWASP_MASVS_L2":[
         "MSTG reference"
      ]
   }
}
"""

PATTERN = "```(.*)```"


@dataclasses.dataclass
class RiskRating(enum.Enum):
    """RiskRating dataclass"""

    INFO = "INFO"
    HARDENING = "HARDENING"
    LOW = "LOW"
    MEDIUM = "MEDIUM"
    HIGH = "HIGH"


@dataclasses.dataclass
class Platform(enum.Enum):
    """Platform dataclass"""

    IOS = "IOS"
    ANDROID = "ANDROID"
    MULTI = "MULTIPLATFORM"
    COMMON = "COMMON"
    WEB = "WEB"


@dataclasses.dataclass
class Vulnerability:
    """Vulnerability dataclass"""

    name: str
    risk_rating: RiskRating
    platform: Platform


@dataclasses.dataclass
class KBEntry:
    """KBEntry dataclass"""

    description: str
    recommendation: str
    meta: dict[str, Any]
    vulnerability: Vulnerability


PLATFORM_TO_PATH = {
    "IOS": "MOBILE_CLIENT/IOS/",
    "ANDROID": "MOBILE_CLIENT/ANDROID/",
    "COMMON": "MOBILE_CLIENT/COMMON/",
    "MULTIPLATFORM": "MOBILE_CLIENT/MULTIPLATFORM_JS/",
    "WEB": "WEB_SERVICE/WEB/",
}


def dump_kb(kbentry: KBEntry) -> None:
    """Dump KB entry into files.

    Args:
        kbentry: Knowledge entry object.
    Returns:

    """
    path_prefix = pathlib.Path(
        PLATFORM_TO_PATH[kbentry.vulnerability.platform.value],
        f"_{kbentry.vulnerability.risk_rating.value}",
    )

    logging.info("KB generated successfully, path is %s", path_prefix)

    path_prefix.mkdir(exist_ok=True, parents=True)

    with pathlib.Path(path_prefix, "description.md").open(
        "w", encoding="utf-8"
    ) as description_md:
        description_md.write(kbentry.description)

    with pathlib.Path(path_prefix, "recommendation.md").open(
        "w", encoding="utf-8"
    ) as recommendation_md:
        recommendation_md.write(kbentry.recommendation)

    with pathlib.Path(path_prefix, "meta.json").open(
        "w", encoding="utf-8"
    ) as meta_json:
        json.dump(kbentry.meta, meta_json, indent=4)


def _ask_gpt(
    prompts: list[dict[str, str]], temperature: float = 0.0, max_tokens: int = 3200
) -> openai_object.OpenAIObject:
    """Send a prompt to OpenAI API."""
    if OPENAI_API_KEY is None:
        raise ValueError

    openai.api_key = OPENAI_API_KEY
    gpt_response: openai_object.OpenAIObject = chat_completion.ChatCompletion.create(
        model=MODEL_NAME,
        temperature=temperature,
        max_tokens=max_tokens,
        messages=prompts,
    )
    return gpt_response


@tenacity.retry(
    stop=tenacity.stop_after_attempt(3),
    wait=tenacity.wait_fixed(2),
    retry=tenacity.retry_if_exception_type(),
)
def generate_kb(vulnerability: Vulnerability) -> KBEntry:
    """Send a prompt to the OpenAI API and generate KB.

    Args:
        vulnerability: a vulnerability object
    Returns:
        KB entry

    """
    description_md = DESCRIPTION_TEMPLATE.replace(
        "%%VULNERABILITY_NAME%%", vulnerability.name.title()
    )
    prompt_message = (
        f"Vulnerability description for {vulnerability.name}, reply as one short paragraph without "
        f"mitigation details "
    )
    prompts = [
        {
            "role": "user",
            "content": prompt_message,
        },
    ]
    gpt_response = _ask_gpt(prompts=prompts)
    content = gpt_response.choices[0].message["content"]
    description_md = description_md.replace("%%VULNERABILITY_DESCRIPTION%%", content)

    recommendation_md = RECOMMENDATION_TEMPLATE.replace(
        "%%VULNERABILITY_NAME%%", vulnerability.name.title()
    )
    prompt_message = f"Vulnerability mitigation for {vulnerability.name}, reply as one short paragraph"
    prompts = [
        {
            "role": "user",
            "content": prompt_message,
        },
    ]
    gpt_response = _ask_gpt(prompts=prompts)
    content = gpt_response.choices[0].message["content"]
    recommendation_md = recommendation_md.replace("%%RECOMMENDATION%%", content)

    for language in DEMO_LANGS:
        prompt_message = (
            f"Demo {language} application that is vulnerable to {vulnerability.name}, "
            "vulnerability has to depend on user input, "
            "application code has to include imports and main function"
            "your response need to consist of the code alone without any extra text"
        )
        prompts = [
            {
                "role": "system",
                "content": "act as a code generator, only reply with code, nothing else",
            },
            {
                "role": "user",
                "content": prompt_message,
            },
        ]
        gpt_response = _ask_gpt(prompts=prompts)
        content = gpt_response.choices[0].message["content"]
        match = re.search(PATTERN, content, re.DOTALL | re.MULTILINE)
        code = match.group(1).strip() if match else "[TODO]"
        description_md = description_md.replace(f"%%{language.upper()}_CODE%%", code)

        prompt_message = (
            f"The {language} code below is vulnerable to {vulnerability.name}, generate a patched version of it:"
            f"{content}"
        )
        prompts = [
            {
                "role": "system",
                "content": "act as a code generator, only reply with code, nothing else",
            },
            {
                "role": "user",
                "content": prompt_message,
            },
        ]
        gpt_response = _ask_gpt(prompts=prompts)
        content = gpt_response.choices[0].message["content"]
        match = re.search(PATTERN, content, re.DOTALL | re.MULTILINE)
        code = match.group(1).strip() if match else "[TODO]"
        recommendation_md = recommendation_md.replace(
            f"%%{language.upper()}_CODE%%", code
        )

    prompt_message = (
        f"Generate a metadata for {vulnerability.name} vulnerability, use the following template: "
        f"{META_TEMPLATE}"
    )
    prompts = [
        {
            "role": "system",
            "content": "act as a json metadata generator, only reply with json, nothing else",
        },
        {
            "role": "user",
            "content": prompt_message,
        },
    ]
    gpt_response = _ask_gpt(prompts=prompts)
    content = gpt_response.choices[0].message["content"]
    meta = json.loads(content)

    kbentry = KBEntry(description_md, recommendation_md, meta, vulnerability)

    return kbentry


@click.command()
@click.option("--name", prompt="Enter vulnerability name", help="Vulnerability name")
@click.option(
    "--risk",
    prompt="Enter risk rating",
    help="Risk rating",
    type=click.Choice([risk.value for risk in RiskRating], case_sensitive=False),
)
@click.option(
    "--platform",
    prompt="Enter platform",
    help="Platform",
    type=click.Choice([platform.value for platform in Platform], case_sensitive=False),
)
def main(name: str, risk: str, platform: str) -> None:
    """
    Entry point of the program.

    This function executes the main logic of the program, including initialization,
    user interactions, and finalization steps. It serves as the starting point for
    running the application.

    Args:
        name: vulnerability name
        risk: vulnerability risk rating
        platform: vulnerability target platform
    Returns:

    """
    vulnerability = Vulnerability(name, RiskRating(risk), Platform(platform))
    kbentry = generate_kb(vulnerability)
    dump_kb(kbentry)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ostorlab/agent/message/proto/v3/asset/store/ios_testflight/ios_testflight.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\nOostorlab/agent/message/proto/v3/asset/store/ios_testflight/ios_testflight.proto\x12:ostorlab.agent.message.proto.v3.asset.store.ios_testflight""\n\x07Message\x12\x17\n\x0f\x61pplication_url\x18\x01 \x02(\t'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR,
    "ostorlab.agent.message.proto.v3.asset.store.ios_testflight.ios_testflight_pb2",
    globals(),
)
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    _MESSAGE._serialized_start = 143
    _MESSAGE._serialized_end = 177
# @@protoc_insertion_point(module_scope)
